# Name:         envisat.py
# Purpose:      Contains Envisat class definition
# Authors:      Asuka Yamakava, Anton Korosov
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html
from dateutil.parser import parse
import struct

import numpy as np
try:
    import scipy.ndimage
except:
    IMPORT_SCIPY = False
else:
    IMPORT_SCIPY = True

from nansat.vrt import VRT
from nansat.geolocation import Geolocation
from nansat.tools import gdal, ogr

from nansat.exceptions import WrongMapperError, NansatReadError

class Envisat(object):
    """Methods/data shared between Envisat mappers

    This class is needed to read awkward N1 format of ENVISAT
    Mostly it support reading of variables from ADS (additional data sets)
    which are TIE_POINTS_ADS (MERIS) or GEOLOCATION_GRID_ADS (ASAR)
    """

    # specific name of geolocation and offsets,
    # dataTypes and units of each dataset
    allADSParams = {
        'MER_': {'name': 'DS_NAME="Tie points ADS              "\n',
                 'width': 71,
                 'list': {"latitude":
                          {
                          "offset": 13,
                          "dataType": gdal.GDT_Int32,
                          "units": "(10)^-6 deg"
                          },
                          "longitude":
                          {
                          "offset": 13+284*1,
                          "dataType": gdal.GDT_Int32,
                          "units": "(10)^-6 deg"
                          },
                          "DME altitude":
                          {
                          "offset": 13+284*2,
                          "dataType": gdal.GDT_Int32,
                          "units": "m"
                          },
                          "DME roughness":
                          {
                          "offset": 13+284*3,
                          "dataType": gdal.GDT_UInt32,
                          "units": "m"
                          },
                          "DME latitude corrections":
                          {
                          "offset": 13+284*4,
                          "dataType": gdal.GDT_Int32,
                          "units": "(10)^-6 deg"
                          },
                          "DME longitude corrections":
                          {
                          "offset": 13+284*5,
                          "dataType": gdal.GDT_Int32,
                          "units": "(10)^-6 deg"
                          },
                          "sun zenith angles":
                          {
                          "offset": 13+284*6,
                          "dataType": gdal.GDT_UInt32,
                          "units": "(10)^-6 deg"
                          },
                          "sun azimuth angles":
                          {
                          "offset": 13+284*7,
                          "dataType": gdal.GDT_Int32,
                          "units": "(10)^-6 deg"
                          },
                          "viewing zenith angles":
                          {
                          "offset": 13+284*8,
                          "dataType": gdal.GDT_UInt32,
                          "units": "(10)^-6 deg"
                          },
                          "viewing azimuth angles":
                          {
                          "offset": 13+284*9,
                          "dataType": gdal.GDT_Int32,
                          "units": "(10)^-6 deg"
                          },
                          "zonal winds":
                          {
                          "offset": 13+284*10+142*0,
                          "dataType": gdal.GDT_Int16,
                          "units": "m*s-1"
                          },
                          "meridional winds":
                          {
                          "offset": 13+284*10+142*1,
                          "dataType": gdal.GDT_Int16,
                          "units": "m*s-1"
                          },
                          "mean sea level pressure":
                          {
                          "offset": 13+284*10+142*2,
                          "dataType": gdal.GDT_UInt16,
                          "units": "hPa"
                          },
                          "total ozone":
                          {
                          "offset": 13+284*10+142*3,
                          "dataType": gdal.GDT_UInt16,
                          "units": "DU"
                          },
                          "relative humidity":
                          {
                          "offset": 13+284*10+142*4,
                          "dataType": gdal.GDT_UInt16,
                          "units": "%"
                          }
                          }},
        'ASA_': {'name': 'DS_NAME="GEOLOCATION GRID ADS        "\n',
                 'width': 11,
                 'list': {"num_lines":
                          {
                          "offset": 13,
                          "dataType": gdal.GDT_Int16,
                          "units": ""
                          },
                          "first_line_samp_numbers":
                          {
                          "offset": 25+11*4*0,
                          "dataType": gdal.GDT_Float32,
                          "units": ""
                          },
                          "first_line_slant_range_times":
                          {
                          "offset": 25+11*4*1,
                          "dataType": gdal.GDT_Float32,
                          "units": "ns"
                          },
                          "first_line_incidence_angle":
                          {
                          "offset": 25+11*4*2,
                          "dataType": gdal.GDT_Float32,
                          "units": "deg"
                          },
                          "first_line_lats":
                          {
                          "offset": 25+11*4*3,
                          "dataType": gdal.GDT_Int32,
                          "units": "(10)^-6 deg"
                          },
                          "first_line_longs":
                          {
                          "offset": 25+11*4*4,
                          "dataType": gdal.GDT_Int32,
                          "units": "(10)^-6 deg"
                          },
                          "last_line_samp_numbers":
                          {
                          "offset": 25+11*4*5+34+11*4*0,
                          "dataType": gdal.GDT_Int32,
                          "units": ""
                          },
                          "last_line_slant_range_times":
                          {
                          "offset": 25+11*4*5+34+11*4*1,
                          "dataType": gdal.GDT_Float32,
                          "units": "ns"
                          },
                          "last_line_incidence_angle":
                          {
                          "offset": 25+11*4*5+34+11*4*2,
                          "dataType": gdal.GDT_Float32,
                          "units": "deg"
                          },
                          "last_line_lats":
                          {
                          "offset": 25+11*4*5+34+11*4*3,
                          "dataType": gdal.GDT_Int32,
                          "units": "(10)^-6 deg"
                          },
                          "last_line_longs":
                          {
                          "offset": 25+11*4*5+34+11*4*4,
                          "dataType": gdal.GDT_Int32,
                          "units": "(10)^-6 deg"
                          }
                          }
                 }}

    # map: GDAL TYPES ==> struct format strings
    structFmt = {gdal.GDT_Int16: ">h",
                 gdal.GDT_UInt16: ">H",
                 gdal.GDT_Int32: ">i",
                 gdal.GDT_UInt32: ">I",
                 gdal.GDT_Float32: ">f"}
    # names of grids with longitude/latitude in ASAR and MERIS ADS
    lonlatNames = {'ASA_': ['first_line_longs', 'first_line_lats'],
                   'MER_': ['longitude', 'latitude']}

    def setup_ads_parameters(self, filename, gdalMetadata):
        """Select set of params and read offset of ADS"""
        if not gdalMetadata or not ('MPH_PRODUCT' in gdalMetadata.keys()):
            raise WrongMapperError

        self.product = gdalMetadata["MPH_PRODUCT"]
        self.iFileName = filename
        self.prodType = gdalMetadata["MPH_PRODUCT"][0:4]
        self.allADSParams = self.allADSParams[self.prodType]
        self.dsOffsetDict = self.read_offset_from_header(
                                                    self.allADSParams['name'])
        self.lonlatNames = self.lonlatNames[self.prodType]

    def _set_envisat_time(self, gdalMetadata):
        """ Get time from metadata, set time to VRT"""
        # set valid time
        self.dataset.SetMetadataItem('time_coverage_start', parse(gdalMetadata["SPH_FIRST_LINE_TIME"]).isoformat())
        self.dataset.SetMetadataItem('time_coverage_end', parse(gdalMetadata["SPH_LAST_LINE_TIME"]).isoformat())


    def read_offset_from_header(self, gadsDSName):
        """ Read offset of ADS from text header.

        Find a location of gadsDSName.
        Adjust the location with textOffset and read the text at the location.
        Convert the text to integer and set it into offsetDict.

        Returns
        -------
            offsetDict : dictionary
                offset of DS, size of DS, number of records, size of record
        """
        # number of lines after line with 'DS_NAME'
        textOffset = {'DS_OFFSET': 3, 'DS_SIZE': 4,
                      'NUM_DSR': 5, 'DSR_SIZE': 6}

        # open file and read 150 header lines
        with open(self.iFileName, "rb") as f:
            headerLines = [f.readline().decode('utf-8') for i in range(150)]

        offsetDict = {}
        # create a dictionary with offset, size, number of records,
        # size of records
        if gadsDSName in headerLines:
            # get location of gadsDSName
            gridOffset = headerLines.index(gadsDSName)
            # Adjust the location of the varaibles by adding textOffset.
            # Read a text at the location and convert the text into integer.
            for iKey in textOffset:
                offsetDict[iKey] = int(headerLines[gridOffset +
                                       textOffset[iKey]].replace(iKey+"=", '').
                                       replace('<bytes>', ''))
        return offsetDict

    def read_binary_line(self, offset, fmtString, length):
        """Read line with binary data at given offset

        Open file
        Read values of given size, at given offset, given times
        Convert to list of values of given format and return

        Parameters
        ----------
            offset: int, start of reading
            fmtString: str, data type format
            lentgh: number of values to read

        Returns
        -------
            binaryValues : list
                values which are read from the file.
                the number of elements is length
        """
        # fseek, read all into a list
        f = open(self.iFileName, 'rb')
        f.seek(offset, 0)
        binaryValues = []
        for i in range(length):
            fString = f.read(struct.calcsize(fmtString))
            fVal = struct.unpack(fmtString, fString)[0]
            binaryValues.append(fVal)
        f.close()

        return binaryValues

    def read_scaling_gads(self, indeces):
        """ Read Scaling Factor GADS to get scalings of MERIS L1/L2

        Parameters
        ----------
            filename : string
            indeces : list

        Returns
        -------
            list
        """
        maxGADS = max(indeces) + 1
        dsOffsetDict = self.read_offset_from_header(
            'DS_NAME="Scaling Factor GADS         "\n')
        allGADSValues = self.read_binary_line(dsOffsetDict["DS_OFFSET"],
                                              '>f', maxGADS)
        #get only values required for the mapper
        return [allGADSValues[i] for i in indeces]

    def get_array_from_ADS(self, adsName):
        """ Create VRT with a band from Envisat ADS metadata

        Read offsets of the <adsName> ADS.
        Read 2D matrix of binary values from ADS from file.
        Read last line ADS (in case of ASAR).
        Zoom array with ADS data to <zoomSize>. Zooming is needed to create
        smooth matrices. Array is zoomed to small size because it is stred in
        memory. Later the VRT with zoomed array is VRT.get_resized_vrt()
        in order to match the size of the Nansat onject.
        Create VRT from the ADS array.

        Parameters
        ----------
            adsName : str
                name of variable from ADS to read. should match allADSParams

        Returns
        -------
            adsVrt : VRT
                vrt with a band created from ADS array

        """
        # Get parameters of arrays in ADS
        adsWidth = self.allADSParams['width']
        adsParams = self.allADSParams['list'][adsName]

        # get data type format string and size
        fmtString = self.structFmt[adsParams['dataType']]

        # create an array whose elements are fetched from ADS
        array = np.array([])

        # read sequence of 1D arrays from ADS
        adsHeight = self.dsOffsetDict["NUM_DSR"]
        for i in range(adsHeight):
            lineOffset = (self.dsOffsetDict['DS_OFFSET'] +
                          adsParams['offset'] +
                          self.dsOffsetDict["DSR_SIZE"] * i)
            binaryLine = self.read_binary_line(lineOffset, fmtString, adsWidth)
            array = np.append(array, binaryLine)

        # read 'last_line_...'
        if self.prodType == 'ASA_':
            adsName = adsName.replace('first_line', 'last_line')
            adsParams = self.allADSParams['list'][adsName]
            lineOffset = (self.dsOffsetDict['DS_OFFSET'] +
                          adsParams['offset'] +
                          self.dsOffsetDict["DSR_SIZE"] * i)
            binaryLine = self.read_binary_line(lineOffset, fmtString, adsWidth)
            array = np.append(array, binaryLine)
            adsHeight += 1

        # adjust the scale
        if '(10)^-6' in adsParams['units']:
            array /= 1000000.0
            # Commenting out line below, otherwise subsequent calls
            # for lon and lat results in modified unit,
            # and hence necessary scaling is not performed
            # adsParams['units'] = adsParams['units'].replace('(10)^-6 ', '')

        # reshape the array into 2D matrix
        array = array.reshape(adsHeight, adsWidth)
        return array

    def create_VRT_from_ADS(self, adsName, zoomSize=500):
        """ Create VRT with a band from Envisat ADS metadata

        Read offsets of the <adsName> ADS.
        Read 2D matrix of binary values from ADS from file.
        Zoom array with ADS data to <zoomSize>. Zooming is needed to create smooth matrices. Array
        is zoomed to small size because it is stred in memory. Later the VRT with zoomed array is
        VRT.get_resized_vrt() in order to match the size of the Nansat object.

        Create VRT from the ADS array.

        Parameters
        ----------
            adsName : str
                name of variable from ADS to read. should match allADSParams
            zoomSize :  int, optional, 500
                size, to which original matrix from ADSR is zoomed using
                scipy.zoom

        Returns
        -------
            adsVrt : VRT, vrt with a band created from ADS array

        """
        adsHeight = self.dsOffsetDict["NUM_DSR"]
        adsParams = self.allADSParams['list'][adsName]
        array = self.get_array_from_ADS(adsName)

        if not IMPORT_SCIPY:
            raise NansatReadError('ENVISAT data cannot be read because scipy is not installed...')

        # zoom the array
        array = scipy.ndimage.interpolation.zoom(array,
                                                 zoomSize / float(adsHeight),
                                                 order=1)

        # create VRT from the array
        adsVrt = VRT.from_array(array=array)
        # add "name" and "units" to band metadata
        bandMetadata = {"name": adsName, "units": adsParams['units']}
        adsVrt.dataset.GetRasterBand(1).SetMetadata(bandMetadata)

        return adsVrt

    def get_ads_vrts(self, gdalDataset, adsNames, zoomSize=500,
                     step=1, **kwargs):
        """Create list with VRTs with zoomed and resized ADS arrays

        For given names of variables (which should match self.allADSParams):
            Get VRT with zoomed ADS array
            Get resized VRT

        Parameters
        ----------
            gdalDataset: GDAL Dataset
                input dataset
            adsNames: list with strings
                names of varaiables from self.allADSParams['list']
            zoomSize: int, 500
                size to which the ADS array will be zoomed by scipy.zoom
            step: int, 1
                step, at which data will be given

        Returns
        --------
            adsVRTs: list with VRT
                list with resized VRT with zoomed arrays

        """
        XSize = gdalDataset.RasterXSize
        YSize = gdalDataset.RasterYSize
        # list with VRT with arrays of lon/lat
        adsVRTs = []
        for adsName in adsNames:
            # create VRT with array from ADS
            adsVRTs.append(self.create_VRT_from_ADS(adsName, zoomSize))
            # resize the VRT to match <step>
            adsVRTs[-1] = adsVRTs[-1].get_resized_vrt(XSize/step, YSize/step)
        return adsVRTs

    def add_geolocation_from_ads(self, gdalDataset, zoomSize=500, step=1):
        """
        Add geolocation domain metadata to the dataset

        Get VRTs with zoomed arrays of lon and lat
        Create geolocation object and add to the metadata

        Parameters
        ----------
        gdalDataset: GDAL Dataset
            input dataset
        zoomSize: int, optional, 500
            size, to which the ADS array will be zoomed using scipy
            array of this size will be stored in memory
        step: int
            step of pixel and line in GeolocationArrays. lat/lon grids are
            generated at that step

        Modifies
        --------
        Adds Geolocation Array metadata
        """
        # get VRTs with lon and lat
        xyVRTs = self.get_ads_vrts(gdalDataset, self.lonlatNames, zoomSize,
                                   step)

        # Add geolocation domain metadata to the dataset
        self._add_geolocation(Geolocation(x_vrt=xyVRTs[0],
                                  y_vrt=xyVRTs[1],
                                  x_band=1, y_band=1,
                                  srs=gdalDataset.GetGCPProjection(),
                                  line_offset=0,
                                  line_step=step,
                                  pixel_offset=0,
                                  pixel_step=step))




# Name:         mapper_emodnet.py
# Purpose:      Mapper for bathymetry data from EMODNet
#               http://portal.emodnet-bathymetry.eu/
# Authors:      Anton Korosov
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html
import os
from dateutil.parser import parse

import numpy as np

from nansat.nsr import NSR
from nansat.vrt import VRT
from nansat.node import Node
from nansat.tools import gdal, ogr
from nansat.exceptions import WrongMapperError


class Mapper(VRT):
    def __init__(self, inputFileName, gdalDataset, gdalMetadata, logLevel=30,
                 **kwargs):
        # check if mapper fits
        if not gdalMetadata:
            raise WrongMapperError
        if not os.path.splitext(inputFileName)[1] == '.mnt':
            raise WrongMapperError
        try:
            mbNorthLatitude = float(gdalMetadata['NC_GLOBAL#mbNorthLatitude'])
            mbSouthLatitude = float(gdalMetadata['NC_GLOBAL#mbSouthLatitude'])
            mbEastLongitude = float(gdalMetadata['NC_GLOBAL#mbEastLongitude'])
            mbWestLongitude = float(gdalMetadata['NC_GLOBAL#mbWestLongitude'])
            mbProj4String = gdalMetadata['NC_GLOBAL#mbProj4String']
            Number_lines = int(gdalMetadata['NC_GLOBAL#Number_lines'])
            Number_columns = int(gdalMetadata['NC_GLOBAL#Number_columns'])
            Element_x_size = float(gdalMetadata['NC_GLOBAL#Element_x_size'])
            Element_y_size = float(gdalMetadata['NC_GLOBAL#Element_y_size'])
        except:
            raise WrongMapperError

        # find subdataset with DEPTH
        subDatasets = gdalDataset.GetSubDatasets()
        dSourceFile = None
        for subDataset in subDatasets:
            if subDataset[0].endswith('.mnt":DEPTH'):
                dSourceFile = subDataset[0]
        if dSourceFile is None:
            raise WrongMapperError
        dSubDataset = gdal.Open(dSourceFile)
        dMetadata = dSubDataset.GetMetadata()

        try:
            scale_factor = dMetadata['DEPTH#scale_factor']
            add_offset = dMetadata['DEPTH#add_offset']
        except:
            raise WrongMapperError

        geoTransform = [mbWestLongitude, Element_x_size, 0,
                        mbNorthLatitude, 0, -Element_y_size]

        # create empty VRT dataset with geolocation only
        self._init_from_dataset_params(Number_columns, Number_lines, geoTransform, NSR(mbProj4String).wkt, metadata=gdalMetadata)

        metaDict = [{'src': {'SourceFilename': dSourceFile,
                             'SourceBand': 1,
                             'ScaleRatio' : scale_factor,
                             'ScaleOffset' : add_offset},
                     'dst': {'wkv': 'depth'}}]

        # add bands with metadata and corresponding values to the empty VRT
        self.create_bands(metaDict)



# Name:        mapper_obpg_l2
# Purpose:     Mapping for L2 data from the OBPG web-site
# Authors:      Anton Korosov
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html
import os
from dateutil.parser import parse
import json

import numpy as np

import pythesint as pti

from nansat.tools import gdal, ogr
from nansat.vrt import VRT
from nansat.nsr import NSR
from nansat.mappers.obpg import OBPGL2BaseClass

from nansat.exceptions import WrongMapperError

class Mapper(OBPGL2BaseClass):
    ''' Mapper for SeaWIFS/MODIS/MERIS/VIIRS L2 data from OBPG in NC4 format
    '''

    def __init__(self, filename, gdalDataset, gdalMetadata,
                 GCP_COUNT=10, **kwargs):
        ''' Create VRT
        Parameters
        ----------
        GCP_COUNT : int
            number of GCPs along each dimention
        '''

        # extension must be .nc
        if os.path.splitext(filename)[1] != '.nc':
            raise WrongMapperError

        # file must contain navigation_data/longitude
        try:
            ds = gdal.Open('HDF5:"%s"://navigation_data/longitude' % filename)
        except RuntimeError:
            raise WrongMapperError
        else:
            dsMetadata = ds.GetMetadata()

        # title value must be known
        if dsMetadata.get('title', '') not in self.titles:
            raise WrongMapperError

        # get geophysical data variables
        subDatasets = gdal.Open(filename).GetSubDatasets()
        metaDict = []
        for subDataset in subDatasets:
            groupName = subDataset[0].split('/')[-2]
            if groupName not in ['geophysical_data', 'navigation_data']:
                continue
            varName = subDataset[0].split('/')[-1]
            subds = gdal.Open(subDataset[0])
            b = subds.GetRasterBand(1)
            bMetadata = b.GetMetadata()

            # set SRC/DST parameters
            metaEntry = {'src': {'SourceFilename': subDataset[0],
                                 'sourceBand': 1,
                                 'DataType': b.DataType},
                         'dst': {'name': varName}}

            # replace datatype for l2_flags
            if varName == 'l2_flags':
                metaEntry['src']['DataType'] = 4
                metaEntry['src']['SourceType'] = 'SimpleSource'

            # set scale if exist
            metaKey = '%s_%s_scale_factor' % (groupName, varName)
            if metaKey in bMetadata:
                metaEntry['src']['ScaleRatio'] = bMetadata[metaKey]

            # set offset if exist
            metaKey = '%s_%s_add_offset' % (groupName, varName)
            if metaKey in bMetadata:
                metaEntry['src']['ScaleOffset'] = bMetadata[metaKey]

            # set standard_name if exists
            metaKey = '%s_%s_standard_name' % (groupName, varName)
            if metaKey in bMetadata:
                metaEntry['dst']['wkv'] = bMetadata[metaKey]

            # set other metadata
            for metaKey in bMetadata:
                newMetaKey = metaKey.replace('%s_%s_' %  (groupName, varName), '')
                if newMetaKey not in ['scale_factor', 'add_offset', 'DIMENSION_LIST', '_FillValue']:
                    metaEntry['dst'][newMetaKey] = bMetadata[metaKey]
            metaDict.append(metaEntry)

        # make GCPs
        # get lat/lon grids
        longitude = gdal.Open('HDF5:"%s"://navigation_data/longitude' % filename).ReadAsArray()
        latitude = gdal.Open('HDF5:"%s"://navigation_data/latitude' % filename).ReadAsArray()
        rasterYSize, rasterXSize = longitude.shape

        step0 = max(1, int(float(latitude.shape[0]) / GCP_COUNT))
        step1 = max(1, int(float(latitude.shape[1]) / GCP_COUNT))

        gcps = []
        k = 0
        center_lon = 0
        center_lat = 0
        for i0 in range(0, latitude.shape[0], step0):
            for i1 in range(0, latitude.shape[1], step1):
                # create GCP with X,Y,pixel,line from lat/lon matrices
                lon = float(longitude[i0, i1])
                lat = float(latitude[i0, i1])

                if (lon >= -180 and lon <= 180 and lat >= -90 and lat <= 90):
                    gcp = gdal.GCP(lon, lat, 0, i1+0.5, i0+0.5)
                    gcps.append(gcp)
                    center_lon += lon
                    center_lat += lat
                    k += 1

        time_coverage_start = dsMetadata['time_coverage_start']
        time_coverage_end = dsMetadata['time_coverage_end']

        # create VRT
        # x_size, y_size, geo_transform, projection, gcps=None, gcp_projection='', **kwargs
        self._init_from_dataset_params(rasterXSize, rasterYSize, (0, 1, 0, rasterYSize, 0, -1),
                                        NSR().wkt, gcps, NSR().wkt)
        # add bands
        self.create_bands(metaDict)

        # reproject GCPs
        center_lon /= k
        center_lat /= k
        srs = '+proj=stere +datum=WGS84 +ellps=WGS84 +lon_0=%f +lat_0=%f +no_defs' % (center_lon, center_lat)
        self.reproject_GCPs(srs)

        ### BAD, BAd, bad ...
        self.dataset.SetProjection(self.dataset.GetGCPProjection())

        # use TPS for reprojection
        self.tps = True

        # add NansenCloud metadata
        self.dataset.SetMetadataItem('time_coverage_start',
                                     str(time_coverage_start))
        self.dataset.SetMetadataItem('time_coverage_end',
                                     str(time_coverage_end))
        self.dataset.SetMetadataItem('source_type', 'Satellite')
        self.dataset.SetMetadataItem('mapper', 'obpg_l2_nc')


        platform  = {'Orbview-2': 'SEASTAR'}.get(dsMetadata.get('platform'),
                                                 dsMetadata.get('platform'))


        mm = pti.get_gcmd_instrument(dsMetadata.get('instrument'))
        ee = pti.get_gcmd_platform(platform)
        self.dataset.SetMetadataItem('instrument', json.dumps(mm))
        self.dataset.SetMetadataItem('platform', json.dumps(ee))



# Name:         mapper_hirlam2nc.py
# Purpose:      Mapper for Hirlam wind data converted from felt to netCDF
# Authors:      Knut-Frode Dagestad
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html
import datetime

from nansat.tools import gdal, ogr
from nansat.exceptions import WrongMapperError
from nansat.vrt import VRT


class Mapper(VRT):
    def __init__(self, filename, gdalDataset, gdalMetadata, logLevel=30,
                 **kwargs):

        if not gdalMetadata:
            raise WrongMapperError

        isHirlam = False
        for key in gdalMetadata.keys():
            if 'creation by fimex from file' in gdalMetadata[key]:
                isHirlam = True

        if not isHirlam:
            raise WrongMapperError

        #GeolocMetaDict = [{'src':
        #        {'SourceFilename': 'NETCDF:"' + filename + '":longitude',
        #         'SourceBand': 1,
        #         'ScaleRatio': 1,
        #         'ScaleOffset': 0},
        #     'dst': {}},
        #           {'src':
        #        {'SourceFilename': 'NETCDF:"' + filename + '":latitude',
        #         'SourceBand': 1,
        #         'ScaleRatio': 1,
        #         'ScaleOffset': 0},
        #     'dst': {}}]

        subDataset = gdal.Open('NETCDF:"' + filename + '":x_wind_10m')
        #self.GeolocVRT = VRT(srcRasterXSize=subDataset.RasterXSize,
        #                srcRasterYSize=subDataset.RasterYSize)
        #self.GeolocVRT.create_bands(GeolocMetaDict)

        #GeolocObject = GeolocationArray(xVRT=self.GeolocVRT,
        #                                yVRT=self.GeolocVRT,
        #    xBand=1, yBand=2,
        #    lineOffset=0, pixelOffset=0,
        #    lineStep=1, pixelStep=1)

        ## create empty VRT dataset with geolocation only
        #VRT.__init__(self, srcRasterXSize = subDataset.RasterXSize,
        #                   srcRasterYSize = subDataset.RasterYSize,
        #                geolocationArray = GeolocObject,
        #                srcProjection = GeolocObject.d['SRS'])
        lon = gdal.Open(
            'NETCDF:"' + filename + '":longitude"').ReadAsArray()
        lat = gdal.Open(
            'NETCDF:"' + filename + '":latitude"').ReadAsArray()
        self._init_from_lonlat(lon, lat)

        # Add bands with wind components
        metaDict = [{'src': {'SourceFilename': ('NETCDF:"' + filename +
                                                '":x_wind_10m'),
                             'NODATA': -32767},
                     'dst': {'name': 'U',
                             'wkv': 'eastward_wind'}},
                    {'src': {'SourceFilename': ('NETCDF:"' + filename +
                                                '":y_wind_10m'),
                             'NODATA': -32767},
                     'dst': {'name': 'V',
                             'wkv': 'northward_wind'}}]

        # Add pixel function with wind speed
        metaDict.append({'src': [{'SourceFilename': ('NETCDF:"' + filename +
                                                     '":x_wind_10m'),
                                  'SourceBand': 1,
                                  'DataType': 6},
                                 {'SourceFilename': ('NETCDF:"' + filename +
                                                     '":y_wind_10m'),
                                  'SourceBand': 1,
                                  'DataType': 6}],
                         'dst': {'wkv': 'wind_speed',
                                 'name': 'windspeed',
                                 'height': '10 m',
                                 'PixelFunctionType': 'UVToMagnitude',
                                 'NODATA': 9999}})

        # add bands with metadata and corresponding values
        # to the empty VRT
        self.create_bands(metaDict)

        # Add valid time
        validTime = datetime.datetime.utcfromtimestamp(
            int(subDataset.GetRasterBand(1).
                GetMetadata()['NETCDF_DIM_time']))
        self.dataset.SetMetadataItem('time_coverage_start', validTime.isoformat())



# Name:        mapper_goci_l1
# Purpose:     Mapping for GOCI-L1B data
# Authors:     Anton Korosov
# Licence:     This file is part of NANSAT. You can redistribute it or modify
#              under the terms of GNU General Public License, v.3
#              http://www.gnu.org/licenses/gpl-3.0.html

from nansat.tools import gdal, ogr
from nansat.exceptions import WrongMapperError
from nansat.vrt import VRT


class Mapper(VRT):
    ''' VRT with mapping of WKV for MODIS Level 1 (QKM, HKM, 1KM) '''

    def __init__(self, filename, gdalDataset, gdalMetadata, **kwargs):
        ''' Create MODIS_L1 VRT '''

        # raise error in case of not GOCI L1B
        try:
            title = gdalMetadata['HDFEOS_POINTS_Scene_Header_Scene_Title']
        except (TypeError, KeyError):
            raise WrongMapperError
        if not title == 'GOCI Level-1B Data':
            raise WrongMapperError

        # set GOCI projection parameters
        lat_0 = gdalMetadata['HDFEOS_POINTS_Map_Projection_Central_Latitude_(parallel)']
        lon_0 = gdalMetadata['HDFEOS_POINTS_Map_Projection_Central_Longitude_(meridian)']
        rasterXSize = int(gdalMetadata['HDFEOS_POINTS_Scene_Header_number_of_columns'].split(' ')[0])
        rasterYSize = int(gdalMetadata['HDFEOS_POINTS_Scene_Header_number_of_rows'].split(' ')[0])
        proj4 = '+proj=ortho +lat_0=%s +lon_0=%s units=m +ellps=WGS84 +datum=WGS84 +no_defs' % (lat_0, lon_0)
        srs = osr.SpatialReference()
        srs.ImportFromProj4(proj4)
        projection = srs.ExportToWkt()
        geoTransform = (-1391500.0, 500.0, 0.0, 1349500.0, 0.0, -500.0)

        # create empty VRT dataset with georeference only
        self._init_from_dataset_params(rasterXSize, rasterYSize, geoTransform, projection)

        # add bands from subdatasets
        subDatasets = gdalDataset.GetSubDatasets()
        metaDict = []
        wavelengths = ['412', '443', '490', '555', '660', '680', '745', '865']
        for subDSI in range(8):
            metaEntry = {'src': {'SourceFilename': subDatasets[subDSI][0],
                                 'SourceBand': 1,
                                 'ScaleRatio': 1e-6},
                         'dst': {'wkv': 'toa_outgoing_spectral_radiance',
                                 'wavelength': wavelengths[subDSI],
                                 'suffix': wavelengths[subDSI]}}
            metaDict.append(metaEntry)

        # add bands with metadata and corresponding values to the empty VRT
        self.create_bands(metaDict)



# Name:    domain.py
# Purpose: Container of NSR class
# Authors:      Anton Korosov
# Created:      01.02.2014
# Copyright:    (c) NERSC 2011 - 2014
# Licence:
# This file is part of NANSAT.
# NANSAT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
# http://www.gnu.org/licenses/gpl-3.0.html
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
from __future__ import absolute_import, unicode_literals

import osr

from nansat.exceptions import NansatProjectionError

class NSR(osr.SpatialReference, object):
    """Nansat Spatial Reference. Overrides constructor of osr.SpatialReference.

    Parameters
    ----------
    srs : 0, PROJ4 or EPSG or WKT or osr.SpatialReference, NSR
        Specifies spatial reference system (SRS)
        PROJ4:
        string with proj4 options [http://trac.osgeo.org/proj/] e.g.:
        '+proj=latlong +datum=WGS84 +ellps=WGS84 +no_defs'
        '+proj=stere +datum=WGS84 +ellps=WGS84 +lat_0=75 +lon_0=0 +no_defs'
        EPSG:
        integer with EPSG number, [http://spatialreference.org/],
        e.g. 4326
        WKT:
        string with Well Know Text of SRS. E.g.:
        'GEOGCS["WGS 84",
            DATUM["WGS_1984",
                SPHEROID["WGS 84",6378137,298.257223563,
                    AUTHORITY["EPSG","7030"]],
                TOWGS84[0,0,0,0,0,0,0],
                AUTHORITY["EPSG","6326"]],
            PRIMEM["Greenwich",0,
                AUTHORITY["EPSG","8901"]],
            UNIT["degree",0.0174532925199433,
                AUTHORITY["EPSG","9108"]],
            AUTHORITY["EPSG","4326"]]'

    See Also
    --------
    [http://www.gdal.org/gdalwarp.html]
    [http://trac.osgeo.org/proj/]
    [http://spatialreference.org/]
    [http://www.gdal.org/ogr/osr_tutorial.html]

    """
    def __init__(self, srs=0):
        """Create Spatial Reference System from input parameter"""
        # create SRS
        osr.SpatialReference.__init__(self)

        # parse input parameters
        status = 1
        if srs is 0:
            # generate default WGS84 SRS
            status = self.ImportFromWkt(osr.SRS_WKT_WGS84)
        elif isinstance(srs, str):
            # parse as proj4 string
            status = self.ImportFromProj4(str(srs))
            if status > 0:
                # parse as WKT string
                status = self.ImportFromWkt(str(srs))
            if status > 0:
                raise NansatProjectionError('Proj4 or WKT (%s) is wrong' % srs)
        # TODO: catch long in python 3
        elif isinstance(srs, int):
            # parse as EPSG code
            status = self.ImportFromEPSG(srs)
            if status > 0:
                raise NansatProjectionError('EPSG %d is wrong' % srs)
        elif type(srs) in [osr.SpatialReference, NSR]:
            # parse from input Spatial Reference
            status = self.ImportFromWkt(srs.ExportToWkt())
            if status > 0:
                raise NansatProjectionError('NSR %s is wrong' % srs)

    @property
    def wkt(self):
        """Well Known Text representation of SRS"""
        return self.ExportToWkt()



#------------------------------------------------------------------------------
# Name:         test_node.py
# Purpose:      Test the Node class
#
# Author:       Aleksander Vines
#
# Created:      2016-02-26
# Last modified:2016-02-26T16:00
# Copyright:    (c) NERSC
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html
#------------------------------------------------------------------------------
from __future__ import absolute_import
import unittest
import os
from . import nansat_test_data as ntd
from nansat.node import Node


class NodeTest(unittest.TestCase):
    def test_creation(self):
        tag = 'Root'
        value = '   Value   '
        anAttr = 'elValue'
        node = Node(tag, value=value, anAttr=anAttr)
        self.assertEqual(node.tag, tag)
        self.assertDictEqual(node.attributes, {'anAttr': anAttr})
        self.assertEqual(node.value, value.strip())

    def test_getAttributeList(self):
        tag = 'Root'
        value = '   Value   '
        anAttr = 'elValue'
        secondAttr = 'Some value'
        finalAttribute = 'A last value'
        node = Node(tag, value=value, anAttr=anAttr, secondAttr=secondAttr,
                    finalAttribute=finalAttribute)
        nameList, valList = node.getAttributeList()
        self.assertIsInstance(nameList, list)
        self.assertIsInstance(valList, list)
        index = valList.index(anAttr)
        self.assertEqual(nameList[index], 'anAttr')
        index = valList.index(secondAttr)
        self.assertEqual(nameList[index], 'secondAttr')
        index = valList.index(finalAttribute)
        self.assertEqual(nameList[index], 'finalAttribute')

    def test_insert(self):
        contents = ('<Element attr="attrValue"><Subnode>testValue</Subnode>'
                    '</Element>')
        root = Node('root')
        root2 = root.insert(contents)
        element = root2.node('Element')
        rawElement = Node.create(contents)
        self.assertEqual(element.xml(), rawElement.xml())

    def test_create(self):
        test_file_element = os.path.join(ntd.test_data_path,
                                         'some_xml_file.xml')
        fileElement = Node.create(test_file_element)
        with open(test_file_element, 'r') as myfile:
            contents = myfile.read().replace('\n', '')
        root = Node('root')
        root = root.insert(contents)
        rawElement = root.children[0]
        self.assertEqual(fileElement.xml(), rawElement.xml())

    def test_delete_attribute(self):
        tag = 'Root'
        value = '   Value   '
        anAttr = 'elValue'
        node = Node(tag, value=value, anAttr=anAttr)
        self.assertIn('anAttr', node.attributes)
        node.delAttribute('anAttr')
        self.assertNotIn('anAttr', node.attributes)

    def test_add_node(self):
        rootTag = 'Root'
        root = Node(rootTag)
        firstLevelTag = 'FirstLevel'
        firstLevel = Node(firstLevelTag)
        root += firstLevel
        self.assertIn(firstLevel, root.children)

    def test_add_nodes(self):
        rootTag = 'Root'
        root = Node(rootTag)
        firstLevelTag = 'FirstLevel'
        firstLevel = Node(firstLevelTag)
        root += firstLevel
        firstLevel2 = Node(firstLevelTag)
        root += firstLevel2
        firstLevel2ndTag = 'FirstLevel2ndTag'
        firstLevel3 = Node(firstLevel2ndTag)
        root = root + firstLevel3
        self.assertIn(firstLevel, root.children)
        self.assertIn(firstLevel2, root.children)
        self.assertIn(firstLevel3, root.children)

    def test_xml(self):
        rootTag = 'Root'
        root = Node(rootTag)
        firstLevelTag = 'FirstLevel'
        firstLevel = Node(firstLevelTag)
        root += firstLevel
        firstLevel2 = Node(firstLevelTag)
        root += firstLevel2
        firstLevel2ndTag = 'FirstLevel2ndTag'
        firstLevel3 = Node(firstLevel2ndTag)
        root += firstLevel3
        self.assertEqual(root.xml(),
                         ('<Root>\n'
                          '  <FirstLevel/>\n'
                          '  <FirstLevel/>\n'
                          '  <FirstLevel2ndTag/>\n'
                          '</Root>\n'),)

    def test_replace_node(self):
        rootTag = 'Root'
        root = Node(rootTag)
        firstLevelTag = 'FirstLevel'
        firstLevel = Node(firstLevelTag)
        root += firstLevel
        firstLevel2 = Node(firstLevelTag)
        root += firstLevel2
        firstLevel2ndTag = 'FirstLevel2ndTag'
        firstLevel3 = Node(firstLevel2ndTag)
        root.replaceNode(firstLevelTag, 1, firstLevel3)
        self.assertIn(firstLevel, root.children)
        self.assertNotIn(firstLevel2, root.children)
        self.assertIn(firstLevel3, root.children)
        self.assertEqual(len(root.children), 2)

    def test_search_node(self):
        rootTag = 'Root'
        root = Node(rootTag)
        firstLevelTag = 'FirstLevel'
        firstLevel = Node(firstLevelTag)
        root += firstLevel
        firstLevel2 = Node(firstLevelTag)
        root += firstLevel2
        firstLevel2ndTag = 'FirstLevel2ndTag'
        firstLevel3 = Node(firstLevel2ndTag)
        root += firstLevel3
        self.assertEqual(root.node(firstLevelTag,0), firstLevel)
        self.assertEqual(root.node(firstLevelTag,1), firstLevel2)



''' Nansat NetCDF-CF mapper

    Check CF-compliance of your files here:
    http://cfconventions.org/compliance-checker.html
'''

import warnings, os, datetime
import numpy as np
import gdal

from dateutil.parser import parse
from netCDF4 import Dataset

from nansat.vrt import VRT
from nansat.nsr import NSR
from nansat.tools import parse_time

from nansat.exceptions import WrongMapperError, NansatMissingProjectionError


class Mapper(VRT):
    """
    """

    def __init__(self, filename, gdal_dataset, gdal_metadata, *args, **kwargs):

        if not filename.endswith('nc'):
            raise WrongMapperError

        self.input_filename = filename

        if not gdal_metadata:
            raise WrongMapperError

        if 'NC_GLOBAL#GDAL_NANSAT_GCPY_000' in list(gdal_metadata.keys()) or \
                'NC_GLOBAL#GDAL_NANSAT_GCPProjection' in list(gdal_metadata.keys()):
            # Probably Nansat generated netcdf of swath data - see issue #192
            raise WrongMapperError

        metadata = VRT._remove_strings_in_metadata_keys(gdal_metadata,
                                                        ['NC_GLOBAL#', 'NANSAT_', 'GDAL_'])

        # Set origin metadata (TODO: agree on keyword...)
        origin = ''
        nans = 'NANSAT'
        if 'origin' in list(metadata.keys()):
            origin = metadata['origin'] + ' '
        for key in list(metadata.keys()):
            if nans in key:
                metadata['origin'] = origin + nans
            # else: Nothing needs to be done, origin stays the same...

        # Check conventions metadata
        if 'Conventions' not in list(metadata.keys()) or 'CF' not in metadata['Conventions']:
            raise WrongMapperError

        # OBS: at this point, generic mapper fails...
        #if metadata.has_key('GCPProjection'):
        #    # Probably Nansat generated netcdf of swath data - see issue #192
        #    raise WrongMapperError

        # Create empty VRT dataset with geo-reference
        self._create_empty(gdal_dataset, metadata)

        # Add bands with metadata and corresponding values to the empty VRT
        self.create_bands(self._band_list(gdal_dataset, metadata, *args, **kwargs))

        # Check size?
        #xsize, ysize = self.ds_size(sub0)

        # Create complex bands from *_real and *_imag bands (the function is in
        # vrt.py)
        self._create_complex_bands(self._get_sub_filenames(gdal_dataset))

        # Set GCMD/DIF compatible metadata if available
        self._set_time_coverage_metadata(metadata)

        # Then add remaining GCMD/DIF compatible metadata in inheriting mappers

    def times(self):
        ''' Get times from time variable

        NOTE: This cannot be done with gdal because the time variable is a
        vector

        '''
        ds = Dataset(self.input_filename)

        # Get datetime object of epoch and time_units string
        time_units = self._time_units(ds=ds)

        # Get all times - consider caching to save time (see
        # nansat/mappers/opendap.py)
        times = ds.variables[self._timevarname(ds=ds)]

        # Create numpy array of np.datetime64 times (provide epoch to save time)
        tt = np.array([self._time_count_to_np_datetime64(tn,
            time_units=time_units) for tn in times])

        return tt

    def _time_units(self, ds=None):
        if not ds:
            ds = Dataset(self.input_filename)
        times = ds.variables[self._timevarname(ds=ds)]
        rt = parse(times.units, fuzzy=True) # This sets timezone to local
        # Remove timezone information from epoch, which defaults to
        # utc (otherwise the timezone should be given in the dataset)
        epoch = datetime.datetime(rt.year, rt.month, rt.day, rt.hour,
                rt.minute, rt.second)
        return epoch, times.units

    def _timevarname(self, ds=None):
        if not ds:
            ds = Dataset(self.input_filename)
        timevarname = 'time'
        try:
            ncvar = ds.variables[timevarname]
        except KeyError:
            for var in ds.variables:
                try:
                    standard_name = ds.variables[var].standard_name
                except:
                    continue
                if standard_name=='time':
                    timevarname = var
                    break
            ncvar = ds.variables[timevarname]
        return timevarname

    def _time_count_to_np_datetime64(self, time_count, time_units=None):
        if not time_units:
            time_units = self._time_units()
        if 'second' in time_units[1]:
            tt = np.datetime64(time_units[0] +
                    datetime.timedelta(seconds=float(time_count)))
        elif 'hour' in time_units[1]:
            tt = np.datetime64(time_units[0] + datetime.timedelta(hours=int(time_count)))
        elif 'day' in time_units[1]:
            tt = np.datetime64(time_units[0] + datetime.timedelta(days=int(time_count)))
        else:
            raise Exception('Check time units..')
        return tt

    def _band_list(self, gdal_dataset, gdal_metadata, netcdf_dim={}, bands=[]):
        ''' Create list of dictionaries mapping source and destination metadata
        of bands that should be added to the Nansat object.

        Parameters
        ----------
        gdal_dataset : gdal.Dataset
            The gdal dataset opened in nansat.py
        gdal_metadata : dict
            Dictionary of global metadata
        netcdf_dim : dict
            Dictionary of desired slice of a multi-dimensional array. Since
            gdal only returns 2D bands, a multi-dimensional array (x,y,z) is
            split into z bands accompanied with metadata information about the
            position of the slice along the z-axis.
        bands : list
            List of desired bands following NetCDF-CF standard names. NOTE:
            some datasets have other bands as well, i.e., of data not yet
            implemented in CF. We may at some point generalize this to provide
            a dict with key name and value, where the key is, e.g.,
            "standard_name" or "metno_name", etc.
        '''

        class ContinueI(Exception):
            pass

        class BreakI(Exception):
            pass

        metadictlist = []
        ds = Dataset(self.input_filename)
        # Pop netcdf_dim item if the dimension is not in the dimension
        # list of the given dataset
        kpop = []
        for key, val in list(netcdf_dim.items()):
            if key not in ds.dimensions.keys():
                kpop.append(key)
        for key in kpop:
            netcdf_dim.pop(key)

        for fn in self._get_sub_filenames(gdal_dataset):
            if ('GEOLOCATION_X_DATASET' in fn or 'longitude' in fn or
                    'GEOLOCATION_Y_DATASET' in fn or 'latitude' in fn):
                continue
            subds = gdal.Open(fn)
            for i in range(subds.RasterCount):
                band_num = i + 1
                band = subds.GetRasterBand(band_num)
                band_metadata = self._clean_band_metadata(band)
                # Keep only desired bands (given in "bands" list)
                try:
                    if bands:
                        if 'standard_name' not in band_metadata.keys():
                            raise ContinueI
                        if not band_metadata['standard_name'] in bands:
                            raise ContinueI
                except ContinueI:
                    continue
                # Keep only desired slices following "netcdf_dim" dictionary
                try:
                    for key, val in list(netcdf_dim.items()):
                        match = [s for s in band_metadata if key in s]
                        if key == 'time' and type(val) == np.datetime64:
                            # Select band directly from given timestamp, and
                            # break the for loop
                            band_num = np.argmin(np.abs(self.times() - val)) + 1
                            # indexing starts on one, not zero...
                            bdict = self._band_dict(fn, band_num, subds)
                            if bdict:
                                metadictlist.append(bdict)
                            raise BreakI
                        if not match or not band_metadata[match[0]]==val:
                            raise ContinueI
                        else:
                            band_metadata[key] = band_metadata.pop(match[0])
                except ContinueI:
                    continue
                except BreakI:
                    break

                # append band with src and dst dictionaries
                bdict = self._band_dict(fn, band_num, subds, band=band,
                        band_metadata=band_metadata)
                if bdict:
                    metadictlist.append(bdict)

        return metadictlist

    def _clean_band_metadata(self, band, remove = ['_Unsigned', 'ScaleRatio',
        'ScaleOffset', 'PixelFunctionType']):

        band_metadata = band.GetMetadata_Dict()
        for key in remove:
            if key in band_metadata:
                band_metadata.pop(key)

        return band_metadata

    def _band_dict(self, subfilename, band_num, subds, band=None, band_metadata=None):
        '''
        subfilename : string
            Name of subdataset file
        '''

        if not band:
            try:
                band = subds.GetRasterBand(band_num)
            except RuntimeError as e:
                if 'illegal band' in e.message.lower():
                    warnings.warn('Skipping band due to GDAL error: %s' %e.message)
                    return {}
                else:
                    raise
        if not band_metadata:
            band_metadata = self._clean_band_metadata(band)

        if 'time_iso_8601' not in list(band_metadata.keys()):
            if self._timevarname() in list(band_metadata.keys()):
                timecountname = self._timevarname()
            else:
                timecountname = 'NETCDF_DIM_'+self._timevarname()
            try:
                band_metadata['time_iso_8601'] = self._time_count_to_np_datetime64(
                    band_metadata[timecountname])
            except KeyError as e:
                # No timing information available for this band - it is
                # probably a constant, such as land area fraction or similar.
                # Then we don't need time for this band...
                warnings.warn(
                        '%s: %s - %s Continuing without time metadata for band %s'
                        %(e.__repr__().split('(')[0], e.message, e.__doc__,
                            band_metadata['NETCDF_VARNAME']))

        # Generate source metadata
        src = {'SourceFilename': subfilename, 'SourceBand': band_num}

        # Set scale ratio
        scaleRatio = band_metadata.get('ScaleRatio',
                band_metadata.get('scale',
                    band_metadata.get('scale_factor', '')))
        if len(scaleRatio) > 0:
            src['ScaleRatio'] = scaleRatio

        # Set scale offset
        scaleOffset = band_metadata.get('ScaleOffset',
                band_metadata.get('offset',
                    band_metadata.get('add_offset', '')))
        if len(scaleOffset) > 0:
            src['ScaleOffset'] = scaleOffset

        # Set data type
        src['DataType'] = band.DataType

        # Generate destination metadata
        # Copy all metadata from input band
        dst = band_metadata
        # Set wkv
        dst['wkv'] = band_metadata.get('standard_name', '')

        # Set band name
        if 'name' in band_metadata:
            bandName = band_metadata['name']
        else:
            # if it doesn't exist get name from NETCDF_VARNAME
            bandName = band_metadata.get('NETCDF_VARNAME', '')
            if len(bandName) == 0:
                bandName = band_metadata.get('dods_variable', '')

            # remove digits added by gdal when exporting to netcdf
            if (len(bandName) > 0 and 'origin' in band_metadata.keys()
                    and 'nansat' in band_metadata['origin'].lower()):
                if bandName[-1:].isdigit():
                    bandName = bandName[:-1]
                if bandName[-1:].isdigit():
                    bandName = bandName[:-1]

        # if still no bandname, create one
        if len(bandName) == 0:
            bandName = 'band_%03d' % band_num

        dst['name'] = bandName

        return {'src': src, 'dst': dst}

    def _create_empty(self, gdal_dataset, gdal_metadata):
        #subfiles = self._get_sub_filenames_with_projection(gdal_dataset)
        try:
            self._create_empty_from_subdatasets(gdal_dataset, gdal_metadata)
        except NansatMissingProjectionError:
            # this happens rarely - a special workaround is done in mapper_quikscat
            warnings.warn('GDAL cannot determine the dataset projection - using Nansat ' \
                    'spatial reference WKT, assuming a regular longitude/latitude grid')
            self._create_empty_with_nansat_spatial_reference_WKT(gdal_dataset, gdal_metadata)

    def _create_empty_from_subdatasets(self, gdal_dataset, metadata):
        """ Create empty vrt dataset with projection but no bands
        """
        no_projection = True
        # Check if the main gdal dataset has projection
        if gdal_dataset.GetProjection():
            sub = gdal_dataset
            no_projection = False
        else:
            # Loop subdatasets and check for projection
            for fn in self._get_sub_filenames(gdal_dataset):
                sub = gdal.Open(fn)
                if sub.GetProjection():
                    no_projection = False
                    break
        if no_projection:
            raise NansatMissingProjectionError
        # Initialise VRT with subdataset containing projection
        self._init_from_gdal_dataset(sub, metadata=metadata)

    def _create_empty_with_nansat_spatial_reference_WKT(self, gdal_dataset, gdal_metadata):
        """ In this case, gdal cannot find the projection of any
            subdatasets. We therefore assume a regular longitude/latitude grid,
            and set the projection to the Nansat Spatial Reference WKT
            [NSR().wkt], using the first subdataset as source
        """
        # NB: if gdal_dataset has subdatasets, gdal_dataset.RasterXSize equals 512,
        # gdal_dataset.RasterYSize equals 512, gdal_dataset.GetGeoTransform returns
        # (1,0,0,0,1,0). This is wrong.
        fn = self._get_sub_filenames(gdal_dataset)
        if  len(fn) == 0:
            subdataset = gdal_dataset
        else:
            subdataset = gdal.Open(fn[0])

        self._init_from_dataset_params(
                    x_size = subdataset.RasterXSize,
                    y_size = subdataset.RasterYSize,
                    geo_transform = subdataset.GetGeoTransform(),
                    projection = NSR().wkt,
                    metadata = gdal_metadata)

    def _set_time_coverage_metadata(self, gdal_metadata):
        ### GET START TIME from METADATA
        time_coverage_start = None
        if 'time_coverage_start' in gdal_metadata:
            time_coverage_start = parse_time(
                                    gdal_metadata['time_coverage_start'])

        ### GET END TIME from METADATA
        time_coverage_end = None
        if 'time_coverage_end' in gdal_metadata:
            time_coverage_end = parse_time(
                                    gdal_metadata['time_coverage_end'])

        # set time_coverage_start if available
        if time_coverage_start is not None:
            self.dataset.SetMetadataItem('time_coverage_start',
                                    time_coverage_start.isoformat())
        # set time_coverage_end if available
        if time_coverage_end is not None:
            self.dataset.SetMetadataItem('time_coverage_end',
                                    time_coverage_end.isoformat())




# Name:        mapper_obpg_l3
# Purpose:     Mapping for L3 data from the OBPG web-site
# Authors:      Anton Korosov
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html

from __future__ import unicode_literals, absolute_import, division, print_function
import datetime
import os.path
import glob

import numpy as np

from nansat.tools import gdal, ogr
from nansat.vrt import VRT
from nansat.nsr import NSR

from nansat.exceptions import WrongMapperError

class Mapper(VRT):
    ''' Mapper for Level-3 Standard Mapped Image from
    http://oceancolor.gsfc.nasa.gov'''

    # detect wkv from metadata 'Parameter'
    param2wkv = {'Chlorophyll a concentration': 'mass_concentration_of_chlorophyll_a_in_sea_water',
                 'Diffuse attenuation coefficient': 'volume_attenuation_coefficient_of_downwelling_'
                                                    'radiative_flux_in_sea_water',
                 'Remote sensing reflectance': 'surface_ratio_of_upwelling_radiance_emerging_from_'
                                               'sea_water_to_downwelling_radiative_flux_in_air',
                 'CDOM Index': 'volume_absorption_coefficient_of_radiative_flux_in_sea_water_due_'
                               'to_dissolved_organic_matter',
                 'Sea Surface Salinity': 'sea_surface_salinity',
                 'Sea Surface Temperature': 'sea_surface_temperature',
                 'Instantaneous Photosynthetically Available Radiation': 'instantaneous_photosynthetically_available_radiation',
                 'Particle backscatter at 443 nm': 'volume_backscattering_coefficient_of_radiative_flux_in_sea_water_due_to_suspended_particles',
                 'Chlorophyll a concentration, Garver-Siegel-Maritorena Model': 'mass_concentration_of_chlorophyll_a_in_sea_water',
                 'Photosynthetically Available Radiation': 'downwelling_photosynthetic_photon_radiance_in_sea_water',
                 'Instantaneous Photosynthetically Available Radiation': 'instantaneous_downwelling_photosynthetic_photon_radiance_in_sea_water',
                 }

    def __init__(self, filename, gdalDataset, gdalMetadata, **kwargs):
        ''' OBPG L3 VRT '''

        try:
            assert 'Level-3 Standard Mapped Image' in gdalMetadata['Title']
        except:
            raise WrongMapperError

        # get list of similar (same date) files in the directory
        iDir, iFile = os.path.split(filename)
        iFileName, iFileExt = os.path.splitext(iFile)
        simFilesMask = os.path.join(iDir, iFileName)
        simFiles = glob.glob(simFilesMask + iFileExt[0:6] + '*')
        #print 'simFilesMask, simFiles', simFilesMask, simFiles

        metaDict = []
        for simFile in simFiles:
            #print 'simFile', simFile
            # open file, get metadata and get parameter name
            simSupDataset = gdal.Open(simFile)
            if simSupDataset is None:
                # skip this similar file
                #print 'No dataset: %s not a supported SMI file' % simFile
                continue
            # get subdatasets from the similar file
            simSubDatasets = simSupDataset.GetSubDatasets()
            if len(simSubDatasets) > 0:
                for simSubDataset in simSubDatasets:
                    #print 'simSubDataset', simSubDataset
                    if 'l3m_data' in simSubDataset[1]:
                        # get SourceFilename from subdataset
                        tmpSourceFilename = simSubDataset[0]
                        break
            else:
                # get SourceFilename from dataset
                tmpSourceFilename = simFile

            # open subdataset with GDAL
            #print 'tmpSourceFilename', tmpSourceFilename
            tmpGdalDataset = gdal.Open(tmpSourceFilename)

            try:
                # get metadata, get 'Parameter'
                tmpGdalMetadata = tmpGdalDataset.GetMetadata()
                simParameter = tmpGdalMetadata['Parameter']
            except:
                print('No parameter: %s not a supported SMI file')
                continue
            else:
                # set params of the similar file
                simSourceFilename = tmpSourceFilename
                simGdalDataset = tmpGdalDataset
                simGdalMetadata = tmpGdalMetadata

            # get WKV from the similar file
            #print 'simParameter', simParameter
            for param in self.param2wkv:
                #print 'param', param
                if param in simParameter:
                    simWKV = self.param2wkv[param]
                    break

            # generate entry to metaDict
            metaEntry = {'src': {'SourceFilename': simSourceFilename,
                                 'SourceBand':  1,
                                 'ScaleRatio': float(simGdalMetadata['Slope']),
                                 'ScaleOffset': float(simGdalMetadata['Intercept'])},
                         'dst': {'wkv': simWKV}}

            # add wavelength and BandName
            if ' at ' in simParameter and ' nm' in simParameter:
                simWavelength = simParameter.split(' at ')[1].split(' nm')[0]
                metaEntry['dst']['suffix'] = simWavelength
                metaEntry['dst']['wavelength'] = simWavelength

            # add band with Rrsw
            metaEntry2 = None
            if simWKV == 'surface_ratio_of_upwelling_radiance_emerging_from_sea_water_to_downwelling_radiative_flux_in_air':
                metaEntry2 = {'src': [metaEntry['src']]}
                metaEntry2['dst'] = {'wkv': 'surface_ratio_of_upwelling_radiance_emerging_from_sea_water_to_downwelling_radiative_flux_in_water',
                                     'suffix': simWavelength,
                                     'wavelength': simWavelength,
                                     'PixelFunctionType': 'NormReflectanceToRemSensReflectance',
                                     }

            # append entry to metaDict
            metaDict.append(metaEntry)
            if metaEntry2 is not None:
                metaDict.append(metaEntry2)

        #get array with data and make 'mask'
        a = simGdalDataset.ReadAsArray()
        mask = np.zeros(a.shape, 'uint8') + 64
        mask[a < -32000] = 1
        self.band_vrts = {'mask': VRT(array=mask)}

        metaDict.append(
            {'src': {'SourceFilename': self.band_vrts['mask'].filename,
                     'SourceBand':  1},
             'dst': {'name': 'mask'}})

        # create empty VRT dataset with geolocation only
        # print 'simGdalMetadata', simGdalMetadata
        latitudeStep = float(simGdalMetadata.
                             get('Latitude Step',
                                 simGdalMetadata.get('Latitude_Step', 1)))
        longitudeStep = float(simGdalMetadata.
                              get('Longitude Step',
                                  simGdalMetadata.get('Longitude_Step', 1)))
        numberOfColumns = int(simGdalMetadata.
                              get('Number of Columns',
                                  simGdalMetadata.get('Number_of_Columns', 1)))
        numberOfLines = int(simGdalMetadata.
                            get('Number of Lines',
                                simGdalMetadata.get('Number_of_Lines', 1)))
        #longitudeStep = float(simGdalMetadata['Longitude Step'])
        # x_size, y_size, geo_transform, projection, gcps=None, gcp_projection='', **kwargs
        self._init_from_dataset_params(numberOfColumns, numberOfLines,
                                (-180.0, longitudeStep, 0.0, 90.0, 0.0, -longitudeStep),
                                NSR().wkt)

        # add bands with metadata and corresponding values to the empty VRT
        self.create_bands(metaDict)

        # Add valid time
        startYear = int(simGdalMetadata.get('Start Year',
                                            simGdalMetadata.
                                            get('Start_Year', 1)))
        startDay = int(simGdalMetadata.get('Start Day',
                                           simGdalMetadata.
                                           get('Start)Day', 1)))
        self.dataset.SetMetadataItem('time_coverage_start',
            (datetime.datetime(startYear, 1, 1) +
                       datetime.timedelta(startDay)).isoformat())



# ------------------------------------------------------------------------------
# Name:         test_nansat.py
# Purpose:      Test the Nansat class
#
# Author:       Morten Wergeland Hansen, Asuka Yamakawa, Anton Korosov
#
# Created:      18.06.2014
# Last modified:24.08.2017 14:00
# Copyright:    (c) NERSC
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html
# ------------------------------------------------------------------------------
from __future__ import unicode_literals, absolute_import

import os
import logging
import unittest
import warnings
import datetime
from mock import patch, PropertyMock, Mock, MagicMock, DEFAULT
import numpy as np

try:
    if 'DISPLAY' not in os.environ:
        import matplotlib; matplotlib.use('Agg')
    import matplotlib
    import matplotlib.pyplot as plt
except ImportError:
    MATPLOTLIB_IS_INSTALLED = False
else:
    MATPLOTLIB_IS_INSTALLED = True

from nansat import Nansat, Domain, NSR
from nansat.tools import gdal
from nansat.warnings import NansatFutureWarning
from nansat.exceptions import NansatGDALError, WrongMapperError, NansatReadError
from nansat.tests.nansat_test_base import NansatTestBase

warnings.simplefilter("always", NansatFutureWarning)
warnings.simplefilter("always", UserWarning)


class NansatTest(NansatTestBase):

    def test_open_gcps(self):
        with warnings.catch_warnings(record=True) as recorded_warnings:
            n = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)

        nansat_warning_raised = False
        for rw in recorded_warnings:
            if rw.category == NansatFutureWarning:
                nansat_warning_raised = True
        self.assertFalse(nansat_warning_raised)

        self.assertEqual(type(n), Nansat)
        self.assertEqual(n.vrt.dataset.GetProjection(), '')
        self.assertTrue((n.vrt.dataset.GetGCPProjection().startswith('GEOGCS["WGS 84",')))
        self.assertEqual(n.vrt.dataset.RasterCount, 3)
        self.assertEqual(n.filename, self.test_file_gcps)
        self.assertIsInstance(n.logger, logging.Logger)
        self.assertEqual(n.name, os.path.split(self.test_file_gcps)[1])
        self.assertEqual(n.path, os.path.split(self.test_file_gcps)[0])

    def test_open_with_filename_warning(self):
        with warnings.catch_warnings(record=True) as w:
            n = Nansat(fileName=self.test_file_gcps, mapperName=self.default_mapper)
            self.assertEqual(w[0].category, NansatFutureWarning)

    def test_open_with_mappername_warning(self):
        with warnings.catch_warnings(record=True) as w:
            n = Nansat(self.test_file_gcps, mapperName=self.default_mapper)
            self.assertEqual(w[0].category, NansatFutureWarning)

    def test_open_with_loglevel_warning(self):
        with warnings.catch_warnings(record=True) as w:
            n = Nansat(self.test_file_gcps, logLevel=30, mapperName=self.default_mapper)
            self.assertEqual(w[0].category, NansatFutureWarning)

    def test_open_with_domain_warning(self):
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        with warnings.catch_warnings(record=True) as w:
            n = Nansat(domain=d)
            self.assertEqual(len(w), 1)
            self.assertIn('Nansat(domain', str(w[0].message))

    def test_get_time_coverage_start_end(self):
        n = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        n.set_metadata('time_coverage_start', '2016-01-20')
        n.set_metadata('time_coverage_end', '2016-01-21')

        self.assertEqual(type(n.time_coverage_start), datetime.datetime)
        self.assertEqual(type(n.time_coverage_end), datetime.datetime)

    def test_from_domain_array(self):
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        n = Nansat.from_domain(d, np.random.randn(500, 500), {'name': 'band1'})

        self.assertEqual(type(n), Nansat)
        self.assertEqual(type(n[1]), np.ndarray)
        self.assertEqual(n.get_metadata('name', 1), 'band1')
        self.assertEqual(n[1].shape, (500, 500))
        self.assertEqual(n.filename, '')
        self.assertIsInstance(n.logger, logging.Logger)
        self.assertEqual(n.name, '')
        self.assertEqual(n.path, '')

    def test_add_band(self):
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        arr = np.random.randn(500, 500)
        n = Nansat.from_domain(d, log_level=40)
        n.add_band(arr, {'name': 'band1'})

        self.assertEqual(type(n), Nansat)
        self.assertEqual(type(n[1]), np.ndarray)
        self.assertEqual(n.get_metadata('name', 1), 'band1')
        self.assertEqual(n[1].shape, (500, 500))

    def test_add_band_twice(self):
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        arr = np.random.randn(500, 500)
        n = Nansat.from_domain(d, log_level=40)
        n.add_band(arr, {'name': 'band1'})
        n.add_band(arr, {'name': 'band2'})

        self.assertEqual(type(n), Nansat)
        self.assertEqual(type(n[1]), np.ndarray)
        self.assertEqual(type(n[2]), np.ndarray)
        self.assertEqual(n.get_metadata('name', 1), 'band1')
        self.assertEqual(n.get_metadata('name', 2), 'band2')
        self.assertEqual(n[1].shape, (500, 500))
        self.assertEqual(n[2].shape, (500, 500))

    def test_add_bands(self):
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        arr = np.random.randn(500, 500)

        n = Nansat.from_domain(d, log_level=40)
        n.add_bands([arr, arr],
                    [{'name': 'band1'}, {'name': 'band2'}])

        self.assertIsInstance(n, Nansat)
        self.assertEqual(n.vrt.vrt.vrt, None)
        self.assertIsInstance(n[1], np.ndarray)
        self.assertIsInstance(n[2], np.ndarray)
        self.assertEqual(n.get_metadata('name', 1), 'band1')
        self.assertEqual(n.get_metadata('name', 2), 'band2')

    def test_add_bands_no_parameter(self):
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        arr = np.random.randn(500, 500)

        n = Nansat.from_domain(d, log_level=40)
        n.add_bands([arr, arr])

        self.assertEqual(type(n), Nansat)
        self.assertEqual(type(n[1]), np.ndarray)
        self.assertEqual(type(n[2]), np.ndarray)

    def test_add_subvrts_only_to_one_nansat(self):
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        arr = np.random.randn(500, 500)

        n1 = Nansat.from_domain(d, log_level=40)
        n2 = Nansat.from_domain(d, log_level=40)
        n1.add_band(arr, {'name': 'band1'})

        self.assertEqual(type(n1.vrt.band_vrts), dict)
        self.assertTrue(len(n1.vrt.band_vrts) > 0)
        self.assertEqual(n2.vrt.band_vrts, {})

    def test_bands(self):
        n = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        bands = n.bands()

        self.assertEqual(type(bands), dict)
        self.assertTrue(1 in bands)
        self.assertTrue('name' in bands[1])
        self.assertEqual(bands[1]['name'], 'L_645')

    def test_has_band_if_name_matches(self):
        n = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        hb = n.has_band('L_645')
        self.assertTrue(hb)

    def test_has_band_if_standard_name_matches(self):
        n = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        hb = n.has_band('surface_upwelling_spectral_radiance_in_air_emerging_from_sea_water')
        self.assertTrue(hb)

    def test_write_fig_tif(self):
        n = Nansat(self.test_file_arctic, mapperName=self.default_mapper)
        tmpfilename = os.path.join(self.tmp_data_path,
                                   'nansat_write_fig_tif.tif')
        n.write_figure(tmpfilename)
        nn = Nansat(tmpfilename, mapperName=self.default_mapper)
        # Asserts that the basic georeference (corners in this case) is still
        # present after opening the image
        self.assertTrue(np.allclose(n.get_corners(), nn.get_corners()))

    def test_resize_eResampleAlg_is_given(self):
        n = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        n.resize(pixelsize=500, eResampleAlg=5)

    def test_resize_by_pixelsize(self):
        n = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        n.resize(pixelsize=500, resample_alg=1)

        self.assertEqual(type(n[1]), np.ndarray)

    def test_resize_by_factor(self):
        n = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        n.resize(0.5, resample_alg=1)

        self.assertEqual(type(n[1]), np.ndarray)

    def test_resize_by_width(self):
        n = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        n.resize(width=100, resample_alg=1)

        self.assertEqual(type(n[1]), np.ndarray)

    def test_resize_by_height(self):
        n = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        n.resize(height=500, resample_alg=1)

        self.assertEqual(type(n[1]), np.ndarray)

    def test_resize_resize(self):
        n = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        n.resize(0.1)
        n.resize(10)
        tmpfilename = os.path.join(self.tmp_data_path,
                                   'nansat_resize_resize.png')
        n.write_figure(tmpfilename, 2, clim='hist')

        self.assertEqual(type(n[1]), np.ndarray)

    def test_resize_complex_alg_average(self):
        n = Nansat(self.test_file_complex, log_level=40, mapper=self.default_mapper)
        with warnings.catch_warnings(record=True) as w:
            n.resize(0.5, resample_alg=-1)
            self.assertEqual(len(w), 1)
            self.assertTrue(issubclass(w[-1].category, UserWarning))
            self.assertIn('The imaginary parts of complex numbers '
                            'are lost when resampling by averaging ', str(w[-1].message))

    def test_resize_complex_alg0(self):
        n = Nansat(self.test_file_complex, log_level=40, mapper=self.default_mapper)
        n.resize(0.5, resample_alg=0)

        self.assertTrue(np.any(n[1].imag != 0))

    def test_resize_complex_alg1(self):
        n = Nansat(self.test_file_complex, log_level=40, mapper=self.default_mapper)
        n.resize(0.5, resample_alg=1)

        self.assertTrue(np.any(n[1].imag != 0))

    def test_resize_complex_alg2(self):
        n = Nansat(self.test_file_complex, log_level=40, mapper=self.default_mapper)
        n.resize(0.5, resample_alg=2)

        self.assertTrue(np.any(n[1].imag != 0))

    def test_resize_complex_alg3(self):
        n = Nansat(self.test_file_complex, log_level=40, mapper=self.default_mapper)
        n.resize(0.5, resample_alg=3)

        self.assertTrue(np.any(n[1].imag != 0))

    def test_resize_complex_alg4(self):
        n = Nansat(self.test_file_complex, log_level=40, mapper=self.default_mapper)
        n.resize(0.5, resample_alg=4)

        self.assertTrue(np.any(n[1].imag != 0))

    def test_get_GDALRasterBand(self):
        n = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        b = n.get_GDALRasterBand(1)
        arr = b.ReadAsArray()

        self.assertEqual(type(b), gdal.Band)
        self.assertEqual(type(arr), np.ndarray)

    def test_get_GDALRasterBand_if_bandID_is_given(self):
        n = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        b = n.get_GDALRasterBand(bandID=1)
        arr = b.ReadAsArray()

        self.assertEqual(type(b), gdal.Band)
        self.assertEqual(type(arr), np.ndarray)

    def test_list_bands_true(self):
        n = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        lb = n.list_bands(True)

        self.assertEqual(lb, None)

    def test_list_bands_false(self):
        n = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        lb = n.list_bands(False)

        self.assertEqual(type(lb), str)

    def test_reproject_domain(self):
        n = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        d = Domain(4326, "-te 27 70 30 72 -ts 500 500")
        n.reproject(d)
        tmpfilename = os.path.join(self.tmp_data_path, 'nansat_reproject_domain.png')
        n.write_figure(tmpfilename, 2, clim='hist')

        self.assertEqual(n.shape(), (500, 500))
        self.assertEqual(type(n[1]), np.ndarray)
        self.assertTrue(n.has_band('swathmask'))

    def test_reproject_domain_if_dstDomain_is_given(self):
        n = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        d = Domain(4326, "-te 27 70 30 72 -ts 500 500")
        n.reproject(dstDomain=d)
        tmpfilename = os.path.join(self.tmp_data_path, 'nansat_reproject_domain.png')
        n.write_figure(tmpfilename, 2, clim='hist')

        self.assertEqual(n.shape(), (500, 500))
        self.assertEqual(type(n[1]), np.ndarray)
        self.assertTrue(n.has_band('swathmask'))

    def test_reproject_domain_if_eRasampleAlg_is_given(self):
        n = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        d = Domain(4326, "-te 27 70 30 72 -ts 500 500")
        n.reproject(d,eResampleAlg=0)
        tmpfilename = os.path.join(self.tmp_data_path, 'nansat_reproject_domain.png')
        n.write_figure(tmpfilename, 2, clim='hist')

        self.assertEqual(n.shape(), (500, 500))
        self.assertEqual(type(n[1]), np.ndarray)
        self.assertTrue(n.has_band('swathmask'))

    @patch.multiple(Nansat, filename=DEFAULT, __init__ = Mock(return_value=None))
    def test_property_fileName(self, filename):
        n = Nansat()
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            fn = n.fileName
            self.assertTrue( len(w) >= 1 )
            categories = [ww.category for ww in w]
            self.assertIn(NansatFutureWarning, categories)

    @patch.object(Nansat, 'get_corners',
                  return_value=(np.array([0, 0, 360, 360]), np.array([90,-90, 90, -90])))
    def test_reproject_domain_if_source_and_destination_domain_span_entire_lons(self, mock_Nansat):
        n = Nansat(self.test_file_arctic, log_level=40, mapper=self.default_mapper)
        d = Domain(4326, "-te -180 180 60 90 -ts 500 500")
        n.reproject(d)
        tmpfilename = os.path.join(self.tmp_data_path, 'nansat_reproject_domain.png')
        n.write_figure(tmpfilename, 2, clim='hist')

        self.assertEqual(n.shape(), (500, 500))
        self.assertEqual(type(n[1]), np.ndarray)
        self.assertTrue(n.has_band('swathmask'))

    def test_reproject_domain_if_tps_is_given(self):
        n = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        d = Domain(4326, "-te 27 70 30 72 -ts 500 500")
        n.reproject(d, tps=False)
        tmpfilename = os.path.join(self.tmp_data_path,
                                   'nansat_reproject_domain.png')
        n.write_figure(tmpfilename, 2, clim='hist')

        self.assertEqual(n.shape(), (500, 500))
        self.assertEqual(type(n[1]), np.ndarray)
        self.assertTrue(n.has_band('swathmask'))

        n = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        d = Domain(4326, "-te 27 70 30 72 -ts 500 500")
        n.reproject(d, tps=True)
        tmpfilename = os.path.join(self.tmp_data_path,
                                   'nansat_reproject_domain.png')
        n.write_figure(tmpfilename, 2, clim='hist')

        self.assertEqual(n.shape(), (500, 500))
        self.assertEqual(type(n[1]), np.ndarray)
        self.assertTrue(n.has_band('swathmask'))

    def test_reproject_of_complex(self):
        """ Should return np.nan in areas out of swath """
        n = Nansat(self.test_file_complex, log_level=40, mapper=self.default_mapper)
        d = Domain(4326, '-te -92.08 26.85 -92.00 26.91 -ts 200 200')
        n.reproject(d)
        b = n[1]

        self.assertTrue(n.has_band('swathmask'))
        self.assertTrue(np.isnan(b[0, 0]))
        self.assertTrue(np.isfinite(b[100, 100]))

    def test_add_band_and_reproject(self):
        """ Should add band and swath mask and return 0 in areas out of swath """
        n = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        d = Domain(4326, "-te 27 70 30 72 -ts 500 500")
        n.add_band(np.ones(n.shape()))
        n.reproject(d)
        b1 = n[1]
        b4 = n[4]

        self.assertTrue(n.has_band('swathmask')) # the added band
        self.assertTrue(n.has_band('swathmask_0000')) # the actual swathmask
        self.assertTrue(b1[0, 0] == 0)
        self.assertTrue(b1[300, 300] > 0)
        self.assertTrue(np.isnan(b4[0, 0]))
        self.assertTrue(b4[300, 300] == 1.)

    def test_reproject_no_addmask(self):
        """ Should not add swath mask and return 0 in areas out of swath """
        n = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        d = Domain(4326, '-te -92.08 26.85 -92.00 26.91 -ts 200 200')
        n.reproject(d, addmask=False)
        b = n[1]

        self.assertTrue(not n.has_band('swathmask'))
        self.assertTrue(np.isfinite(b[0, 0]))
        self.assertTrue(np.isfinite(b[100, 100]))

    def test_reproject_stere(self):
        n1 = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        n2 = Nansat(self.test_file_stere, log_level=40, mapper=self.default_mapper)
        n1.reproject(n2)
        tmpfilename = os.path.join(self.tmp_data_path,
                                   'nansat_reproject_stere.png')
        n1.write_figure(tmpfilename, 2, clim='hist')

        self.assertEqual(n1.shape(), n2.shape())
        self.assertEqual(type(n1[1]), np.ndarray)

    def test_reproject_gcps(self):
        n1 = Nansat(self.test_file_stere, log_level=40, mapper=self.default_mapper)
        n2 = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        n1.reproject(n2)
        tmpfilename = os.path.join(self.tmp_data_path,
                                   'nansat_reproject_gcps.png')
        n1.write_figure(tmpfilename, 2, clim='hist')

        self.assertEqual(n1.shape(), n2.shape())
        self.assertEqual(type(n1[1]), np.ndarray)

    def test_reproject_gcps_on_repro_gcps(self):
        n1 = Nansat(self.test_file_stere, log_level=40, mapper=self.default_mapper)
        n2 = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        n2.reproject_gcps()
        n1.reproject(n2)
        tmpfilename = os.path.join(self.tmp_data_path,
                                   'nansat_reproject_gcps_on_repro_gcps.png')
        n1.write_figure(tmpfilename, 2, clim='hist')

        self.assertEqual(n1.shape(), n2.shape())
        self.assertEqual(type(n1[1]), np.ndarray)

    def test_reproject_gcps_resize(self):
        n1 = Nansat(self.test_file_stere, log_level=40, mapper=self.default_mapper)
        n2 = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        n1.reproject(n2)
        n1.resize(2)
        tmpfilename = os.path.join(self.tmp_data_path,
                                   'nansat_reproject_gcps_resize.png')
        n1.write_figure(tmpfilename, 2, clim='hist')

        self.assertEqual(n1.shape()[0], n2.shape()[0] * 2)
        self.assertEqual(n1.shape()[1], n2.shape()[1] * 2)
        self.assertEqual(type(n1[1]), np.ndarray)

    def test_undo(self):
        n1 = Nansat(self.test_file_stere, log_level=40, mapper=self.default_mapper)
        shape1 = n1.shape()
        n1.resize(10)
        n1.undo()
        shape2 = n1.shape()

        self.assertEqual(shape1, shape2)

    def test_write_figure(self):
        n1 = Nansat(self.test_file_stere, log_level=40, mapper=self.default_mapper)
        tmpfilename = os.path.join(self.tmp_data_path, 'nansat_write_figure.png')
        n1.write_figure(tmpfilename)

        self.assertTrue(os.path.exists(tmpfilename))

    def test_write_figure_band(self):
        n1 = Nansat(self.test_file_stere, log_level=40, mapper=self.default_mapper)
        tmpfilename = os.path.join(self.tmp_data_path, 'nansat_write_figure_band.png')
        n1.write_figure(tmpfilename, 2)

        self.assertTrue(os.path.exists(tmpfilename))

    def test_write_figure_clim(self):
        n1 = Nansat(self.test_file_stere, log_level=40, mapper=self.default_mapper)
        tmpfilename = os.path.join(self.tmp_data_path, 'nansat_write_figure_clim.png')
        n1.write_figure(tmpfilename, 3, clim='hist')

        self.assertTrue(os.path.exists(tmpfilename))

    def test_write_figure_legend(self):
        n1 = Nansat(self.test_file_stere, log_level=40, mapper=self.default_mapper)
        tmpfilename = os.path.join(self.tmp_data_path, 'nansat_write_figure_legend.png')
        n1.write_figure(tmpfilename, 3, clim='hist', legend=True)

        self.assertTrue(os.path.exists(tmpfilename))

    def test_write_figure_logo(self):
        n1 = Nansat(self.test_file_stere, log_level=40, mapper=self.default_mapper)
        tmpfilename = os.path.join(self.tmp_data_path, 'nansat_write_figure_logo.png')
        n1.write_figure(tmpfilename, 3, clim='hist',
                        logoFileName=self.test_file_gcps)

        self.assertTrue(os.path.exists(tmpfilename))

    def test_write_geotiffimage(self):
        n1 = Nansat(self.test_file_stere, log_level=40, mapper=self.default_mapper)
        tmpfilename = os.path.join(self.tmp_data_path, 'nansat_write_geotiffimage.tif')
        n1.write_geotiffimage(tmpfilename)

        self.assertTrue(os.path.exists(tmpfilename))

    def test_write_geotiffimage_if_bandID_is_given(self):
        n1 = Nansat(self.test_file_stere, log_level=40, mapper=self.default_mapper)
        tmpfilename = os.path.join(self.tmp_data_path, 'nansat_write_geotiffimage.tif')
        n1.write_geotiffimage(tmpfilename, bandID=1)

        self.assertTrue(os.path.exists(tmpfilename))

    def test_get_metadata(self):
        n1 = Nansat(self.test_file_stere, log_level=40, mapper=self.default_mapper)
        m = n1.get_metadata()

        self.assertEqual(type(m), dict)
        self.assertTrue('filename' in m)

    def test_get_metadata_key(self):
        n1 = Nansat(self.test_file_stere, log_level=40, mapper=self.default_mapper)
        m = n1.get_metadata('filename')

        self.assertEqual(type(m), str)

    def test_get_metadata_wrong_key(self):
        n1 = Nansat(self.test_file_stere, log_level=40, mapper=self.default_mapper)

        with self.assertRaises(ValueError):
            n1.get_metadata('some_crap')

    def test_get_metadata_band_id(self):
        n1 = Nansat(self.test_file_stere, log_level=40, mapper=self.default_mapper)
        m = n1.get_metadata(band_id=1)

        self.assertEqual(type(m), dict)
        self.assertTrue('name' in m)

    def test_get_metadata_bandID(self):
        n1 = Nansat(self.test_file_stere, log_level=40, mapper=self.default_mapper)
        m = n1.get_metadata(bandID=1)

        self.assertEqual(type(m), dict)
        self.assertTrue('name' in m)

    def test_set_metadata(self):
        n1 = Nansat(self.test_file_stere, log_level=40, mapper=self.default_mapper)
        n1.set_metadata('newKey', 'newVal')
        m = n1.get_metadata('newKey')

        self.assertEqual(m, 'newVal')

    def test_set_metadata_band_id(self):
        n1 = Nansat(self.test_file_stere, log_level=40, mapper=self.default_mapper)
        n1.set_metadata('newKey', 'newVal', band_id=1)
        m = n1.get_metadata('newKey', 1)

        self.assertEqual(m, 'newVal')

    def test_set_metadata_bandID(self):
        n1 = Nansat(self.test_file_stere, log_level=40, mapper=self.default_mapper)
        n1.set_metadata('newKey', 'newVal', bandID=1)
        m = n1.get_metadata('newKey', 1)

        self.assertEqual(m, 'newVal')

    def test_get_band_number(self):
        n1 = Nansat(self.test_file_stere, log_level=40, mapper=self.default_mapper)
        self.assertEqual(n1._get_band_number(1), 1)

    @unittest.skipUnless(MATPLOTLIB_IS_INSTALLED, 'Matplotlib is required')
    def test_get_transect(self):
        plt.switch_backend('agg')
        n1 = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        t = n1.get_transect([[28.31299128, 28.93691525],
                             [70.93709219, 70.69646524]],
                            [str('L_645')])
        tmpfilename = os.path.join(self.tmp_data_path, 'nansat_get_transect.png')
        plt.plot(t['lat'], t['L_645'], '.-')
        plt.savefig(tmpfilename)
        plt.close('all')

        self.assertTrue('L_645' in t.dtype.fields)
        self.assertTrue('line' in t.dtype.fields)
        self.assertTrue('pixel' in t.dtype.fields)
        self.assertTrue('lat' in t.dtype.fields)
        self.assertTrue('lon' in t.dtype.fields)
        self.assertEqual(type(t['lat']), np.ndarray)
        self.assertEqual(type(t['lon']), np.ndarray)

    def test_get_transect_outside(self):
        n1 = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        t = n1.get_transect([[0, 28.31299128], [0, 70.93709219]], [1])

        self.assertTrue('L_645' in t.dtype.fields)
        self.assertTrue('line' in t.dtype.fields)
        self.assertTrue('pixel' in t.dtype.fields)
        self.assertTrue('lat' in t.dtype.fields)
        self.assertTrue('lon' in t.dtype.fields)
        self.assertEqual(type(t['lat']), np.ndarray)
        self.assertEqual(type(t['lon']), np.ndarray)

    def test_get_transect_wrong_points(self):
        n1 = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        self.assertRaises(ValueError, n1.get_transect, [1, 1], [1])

    def test_get_transect_wrong_band(self):
        n1 = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        t = n1.get_transect([[0, 28.31299128], [0, 70.93709219]], [10])

        self.assertTrue('line' in t.dtype.fields)
        self.assertTrue('pixel' in t.dtype.fields)
        self.assertTrue('lat' in t.dtype.fields)
        self.assertTrue('lon' in t.dtype.fields)
        self.assertEqual(type(t['lat']), np.ndarray)
        self.assertEqual(type(t['lon']), np.ndarray)

    def test_get_transect_pixlin(self):
        n1 = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        t = n1.get_transect([[10, 20],
                             [10, 10]],
                             [str('L_645')],
                            lonlat=False)

        self.assertTrue('L_645' in t.dtype.fields)
        self.assertTrue('line' in t.dtype.fields)
        self.assertTrue('pixel' in t.dtype.fields)
        self.assertTrue('lat' in t.dtype.fields)
        self.assertTrue('lon' in t.dtype.fields)
        self.assertEqual(type(t['lat']), np.ndarray)
        self.assertEqual(type(t['lon']), np.ndarray)
        self.assertEqual(len(t['lon']), 11)

    def test_get_transect_data(self):
        n1 = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        b1 = n1[1]
        t = n1.get_transect([[28.3], [70.9]], [], data=b1)

        self.assertTrue('input' in t.dtype.fields)
        self.assertTrue('L_645' not in t.dtype.fields)
        self.assertTrue('line' in t.dtype.fields)
        self.assertTrue('pixel' in t.dtype.fields)
        self.assertTrue('lat' in t.dtype.fields)
        self.assertTrue('lon' in t.dtype.fields)
        self.assertEqual(type(t['lat']), np.ndarray)
        self.assertEqual(type(t['lon']), np.ndarray)

    @patch('nansat.nansat.PointBrowser')
    def test_digitize_points(self, mock_PointBrowser):
        """ shall create PointBrowser and call PointBrowser.get_points() """
        value = 'points'
        mock_PointBrowser().get_points.return_value = value
        n = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        points = n.digitize_points(1)
        self.assertTrue(mock_PointBrowser.called_once())
        self.assertEqual(points, value)

    def test_crop(self):
        n1 = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        ext = n1.crop(10, 20, 50, 60)

        self.assertEqual(n1.shape(), (60, 50))
        self.assertEqual(ext, (10, 20, 50, 60))
        self.assertEqual(type(n1[1]), np.ndarray)

        n1 = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        ext = n1.crop(0, 0, 200, 200)

        self.assertEqual(n1.shape(), (200, 200))
        self.assertEqual(ext, (0, 0, 200, 200))
        self.assertEqual(type(n1[1]), np.ndarray)


    def test_crop_gcpproj(self):
        n1 = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        n1.reproject_gcps()
        ext = n1.crop(10, 20, 50, 60)
        xmed = abs(np.median(np.array([gcp.GCPX
                                for gcp in n1.vrt.dataset.GetGCPs()])))
        gcpproj = NSR(n1.vrt.dataset.GetGCPProjection()
                                        ).ExportToProj4().split(' ')[0]

        self.assertTrue(xmed > 360)
        self.assertTrue(gcpproj=='+proj=stere')

    def test_crop_complex(self):
        n1 = Nansat(self.test_file_complex, log_level=40, mapper=self.default_mapper)
        ext = n1.crop(10, 20, 50, 60)

        self.assertEqual(n1.shape(), (60, 50))
        self.assertEqual(ext, (10, 20, 50, 60))
        self.assertEqual(type(n1[1]), np.ndarray)

    def test_crop_no_gcps_arctic(self):
        n1 = Nansat(self.test_file_arctic, log_level=40, mapper=self.default_mapper)
        ext = n1.crop(10, 20, 50, 60)

        self.assertEqual(n1.shape(), (60, 50))
        self.assertEqual(ext, (10, 20, 50, 60))
        self.assertEqual(type(n1[1]), np.ndarray)

    def test_crop_lonlat(self):
        n1 = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        ext = n1.crop_lonlat([28, 29], [70.5, 71])

        self.assertEqual(n1.shape(), (111, 110))
        self.assertEqual(ext, (31, 89, 110, 111))
        self.assertEqual(type(n1[1]), np.ndarray)

    def test_crop_outside(self):
        n1 = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        self.assertRaises(ValueError, n1.crop_lonlat, [-10, 10], [-10, 10])

    def test_watermask(self):
        """ if watermask data exists: should fetch array with watermask
            else:                     should raise an error """
        n1 = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        mod44path = os.getenv('MOD44WPATH')
        if mod44path is not None and os.path.exists(mod44path + '/MOD44W.vrt'):
            wm = n1.watermask()[1]
            self.assertEqual(type(wm), np.ndarray)
            self.assertEqual(wm.shape[0], n1.shape()[0])
            self.assertEqual(wm.shape[1], n1.shape()[1])

    def test_watermask_fail_if_mod44path_is_wrong(self):
        """ Nansat.watermask should raise an IOError"""
        n1 = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        os.environ['MOD44WPATH'] = '/fakepath'
        self.assertRaises(IOError, n1.watermask)

    def test_watermask_fail_if_mod44path_not_exist(self):
        """ Nansat.watermask should raise an IOError"""
        n1 = Nansat(self.test_file_gcps, log_level=40, mapper=self.default_mapper)
        del os.environ['MOD44WPATH']
        self.assertRaises(IOError, n1.watermask)

    def test_init_no_arguments(self):
        """ No arguments should raise ValueError """
        self.assertRaises(ValueError, Nansat)

    def test_get_item_basic_expressions(self):
        """ Testing get_item with some basic expressions """
        self.mock_pti['get_wkv_variable'].return_value=dict(short_name='newband')
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        n = Nansat.from_domain(d, np.zeros((500, 500)), {'expression': 'np.ones((500, 500))'})
        self.assertIsInstance(n[1], np.ndarray)
        self.assertEqual(n[1].shape, (500, 500))
        band1 = n[1]
        self.assertTrue(np.allclose(band1, np.ones((500, 500))))

    def test_get_item_inf_expressions(self):
        """ inf should be replaced with nan """
        self.mock_pti['get_wkv_variable'].return_value=dict(short_name='newband')
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        n = Nansat.from_domain(d, log_level=40)
        arr = np.empty((500, 500))
        n.add_band(arr, {'expression': 'np.array([0,1,2,3,np.inf,5,6,7])'})
        self.assertIsInstance(n[1], np.ndarray)
        self.assertTrue(np.isnan(n[1][4]))

    def test_repr_basic(self):
        """ repr should include some basic elements """
        d = Domain(4326, "-te 25 70 35 72 -ts 500 500")
        n = Nansat.from_domain(d, log_level=40)
        arr = np.empty((500, 500))
        exp = 'np.array([0,1,2,3,np.inf,5,6,7])'
        n.add_band(arr, {'expression': exp})
        n_repr = repr(n)
        self.assertIn(exp, n_repr, 'The expressions should be in repr')
        self.assertIn('SourceFilename', n_repr)
        self.assertIn('/vsimem/', n_repr)
        self.assertIn('500 x 500', n_repr)
        self.assertIn('Projection:', n_repr)
        self.assertIn('25', n_repr)
        self.assertIn('72', n_repr)
        self.assertIn('35', n_repr)
        self.assertIn('70', n_repr)

    @patch.object(Nansat, 'get_GDALRasterBand')
    def test_getitem(self, mock_Nansat):
        type(mock_Nansat()).GetMetadata = MagicMock(return_value={'a':1})
        type(mock_Nansat()).ReadAsArray = MagicMock(return_value=None)
        with self.assertRaises(NansatGDALError):
            Nansat(self.test_file_stere, mapperName=self.default_mapper).__getitem__(1)

    @patch.object(Nansat, 'digitize_points')
    def test_crop_interactive(self, mock_digitize_points):
        mock_digitize_points.return_value=[np.array([[10, 20], [10, 30]])]
        n = Nansat(self.test_file_arctic, log_level=40, mapper=self.default_mapper)
        n.crop_interactive()
        self.assertEqual(n.shape(), (20, 10))

    def test_extend(self):
        n = Nansat(self.test_file_arctic, log_level=40, mapper=self.default_mapper)
        nshape1 = n.shape()
        n.extend(left=10, right=20, top=30, bottom=40)
        be = n[1]
        self.assertEqual(n.shape(), (nshape1[0]+70, nshape1[1]+30))
        self.assertIsInstance(be, np.ndarray)

    def test_open_no_mapper(self):
        n = Nansat(self.test_file_arctic)
        self.assertEqual(type(n), Nansat)
        self.assertEqual(n.mapper, 'netcdf_cf')

if __name__ == "__main__":
    unittest.main()



# Name:        mapper_modisL1
# Purpose:     Mapping for MODIS-L1 data
# Authors:      Anton Korosov
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html
from __future__ import absolute_import, unicode_literals, division

import os
import glob
from datetime import datetime, timedelta
from math import ceil
try:
    from scipy.ndimage.filters import gaussian_filter
except:
    IMPORT_SCIPY = False
else:
    IMPORT_SCIPY = True

from nansat.nsr import NSR
from nansat.vrt import VRT
from nansat.tools import gdal, ogr
from nansat.exceptions import WrongMapperError, NansatReadError


class Mapper(VRT):
    ''' VRT with mapping of WKV for VIIRS Level 1B '''

    def __init__(self, filename, gdalDataset, gdalMetadata,
                 GCP_COUNT0=5, GCP_COUNT1=20, pixelStep=1,
                 lineStep=1, **kwargs):
        ''' Create VIIRS VRT '''

        if not 'GMTCO_npp_' in filename:
            raise WrongMapperError(filename)
        ifiledir = os.path.split(filename)[0]
        ifiles = glob.glob(ifiledir + 'SVM??_npp_d*_obpg_ops.h5')
        ifiles.sort()

        if not IMPORT_SCIPY:
            raise NansatReadError('VIIRS data cannot be read because scipy is not installed')

        viirsWavelengths = [None, 412, 445, 488, 555, 672, 746, 865, 1240,
                            1378, 1610, 2250, 3700, 4050, 8550, 10736, 12013]

        # create empty VRT dataset with geolocation only
        xDatasetSource = ('HDF5:"%s"://All_Data/VIIRS-MOD-GEO-TC_All/Longitude'
                          % filename)
        xDatasetBand = 1
        xDataset = gdal.Open(xDatasetSource)
        self._init_from_gdal_dataset(xDataset)

        metaDict = []
        for ifile in ifiles:
            ifilename = os.path.split(ifile)[1]
            print(ifilename)
            bNumber = int(ifilename[3:5])
            print(bNumber)
            bWavelength = viirsWavelengths[bNumber]
            print(bWavelength)
            SourceFilename = ('HDF5:"%s"://All_Data/VIIRS-M%d-SDR_All/Radiance'
                              % (ifile, bNumber))
            print(SourceFilename)
            metaEntry = {'src': {'SourceFilename': SourceFilename,
                         'SourceBand': 1},
                         'dst': {'wkv': 'toa_outgoing_spectral_radiance',
                                 'wavelength': str(bWavelength),
                                 'suffix': str(bWavelength)}
                         }
            metaDict.append(metaEntry)

        # add bands with metadata and corresponding values to the empty VRT
        self.create_bands(metaDict)

        xVRTArray = xDataset.ReadAsArray()
        xVRTArray = gaussian_filter(xVRTArray, 5).astype('float32')
        xVRT = VRT.from_array(xVRTArray)

        yDatasetSource = ('HDF5:"%s"://All_Data/VIIRS-MOD-GEO-TC_All/Latitude'
                          % filename)
        yDatasetBand = 1
        yDataset = gdal.Open(yDatasetSource)
        yVRTArray = yDataset.ReadAsArray()
        yVRTArray = gaussian_filter(yVRTArray, 5).astype('float32')
        yVRT = VRT.from_array(yVRTArray)

        # estimate pixel/line step
        self.logger.debug('pixel/lineStep %f %f' % (pixelStep, lineStep))

        # ==== ADD GCPs and Pojection ====
        # get lat/lon matrices
        longitude = xVRT.dataset.GetRasterBand(1).ReadAsArray()
        latitude = yVRT.dataset.GetRasterBand(1).ReadAsArray()

        # estimate step of GCPs
        step0 = max(1, int(float(latitude.shape[0]) / GCP_COUNT0))
        step1 = max(1, int(float(latitude.shape[1]) / GCP_COUNT1))
        self.logger.debug('gcpCount: %d %d %d %d, %d %d',
                          latitude.shape[0], latitude.shape[1],
                          GCP_COUNT0, GCP_COUNT1, step0, step1)

        # generate list of GCPs
        gcps = []
        k = 0
        for i0 in range(0, latitude.shape[0], step0):
            for i1 in range(0, latitude.shape[1], step1):
                # create GCP with X,Y,pixel,line from lat/lon matrices
                lon = float(longitude[i0, i1])
                lat = float(latitude[i0, i1])
                if (lon >= -180 and lon <= 180 and lat >= -90 and lat <= 90):
                    gcp = gdal.GCP(lon, lat, 0,
                                   i1 * pixelStep, i0 * lineStep)
                    self.logger.debug('%d %d %d %f %f',
                                      k, gcp.GCPPixel, gcp.GCPLine,
                                      gcp.GCPX, gcp.GCPY)
                    gcps.append(gcp)
                    k += 1

        # append GCPs and lat/lon projection to the vsiDataset
        self.dataset.SetGCPs(gcps, NSR().wkt)

        # remove geolocation array
        self._remove_geolocation()
        #"""



import unittest 

import os, datetime
import numpy as np

from nansat.nansat import Nansat

from nansat.tests import nansat_test_data as ntd

class NetCDFCFMapperTests(unittest.TestCase):

    def setUp(self):
        self.test_file_arctic = os.path.join(ntd.test_data_path, 'arctic.nc')
        self.test_file_arome_metcoop = '/vagrant/shared/test_data/generic/arome_metcoop_default2_5km_20170227T00Z.nc'
        self.test_file_arome_arctic = '/vagrant/shared/test_data/generic/arome_arctic_pp_2_5km_20170227T00Z.nc'
        self.test_file_ecmwf = '/vagrant/shared/test_data/generic/ec_atmo_0_1deg_20170227T000000Z_1h.nc'
        self.test_file_arome_opendap = 'http://thredds.met.no/thredds/catalog/arome25/catalog.html?dataset=arome25/arome_metcoop_default2_5km_latest.nc'
        self.s1aEW = '/vagrant/shared/test_data/sentinel1_l1/S1A_EW_GRDM_1SDH_20170227T065537_20170227T065637_015466_019652_DD9C.SAFE'
        self.s1bIW = '/vagrant/shared/test_data/sentinel1_l1/S1B_IW_GRDM_1SDV_20170227T061040_20170227T061113_004482_007CD7_3205.SAFE'

    def test_netcdf_cf_mapper_is_used(self):
        n = Nansat(self.test_file_arctic)
        self.assertEqual(n.mapper, 'netcdf_cf')

    def test_open_netcdf_cf(self):
        n = Nansat(self.test_file_arctic, mapperName='netcdf_cf')
        self.assertIsInstance(n, Nansat)

    def test_open_arome_metcoop(self):
        n = Nansat(self.test_file_arome_metcoop, mapperName='netcdf_cf')
        self.assertIsInstance(n, Nansat)
        self.assertTrue(n['x_wind_10m'].any())
        self.assertTrue(n['y_wind_10m'].any())

    def test_open_arome_metcoop_at_given_time(self):
        n = Nansat(self.test_file_arome_metcoop, netcdf_dim={'time': '1488153600'},
                mapperName='netcdf_cf')
        self.assertIsInstance(n, Nansat)
        self.assertTrue(n['x_wind_10m'].any())
        self.assertTrue(n['y_wind_10m'].any())

    def test_open_arome_metcoop_at_given_height(self):
        n = Nansat(self.test_file_arome_metcoop, netcdf_dim={'height0': '0'},
                mapperName='netcdf_cf')
        self.assertIsInstance(n, Nansat)
        self.assertTrue(n['surface_air_pressure'].any())

    def test_open_arome_metcoop_y_wind_and_x_wind_at_given_time(self):
        n = Nansat(self.test_file_arome_metcoop, bands=['y_wind', 'x_wind'],
                netcdf_dim={'time': '1488153600'}, mapperName='netcdf_cf')
        self.assertIsInstance(n, Nansat)
        self.assertEqual(2, len(n.bands()))
        self.assertTrue(n['x_wind_10m'].any())
        self.assertTrue(n['y_wind_10m'].any())

    def test_open_arome_arctic_y_wind_and_x_wind_at_given_time(self):
        n = Nansat(self.test_file_arome_arctic, bands=['y_wind', 'x_wind'],
                netcdf_dim={'time': '1488387600'}, mapperName='netcdf_cf')
        self.assertIsInstance(n, Nansat)
        self.assertEqual(2, len(n.bands()))
        self.assertTrue(n['x_wind_10m'].any())
        self.assertTrue(n['y_wind_10m'].any())

    def test_open_arome_arctic_y_wind_and_x_wind_at_given_datetime(self):
        n = Nansat(self.test_file_arome_arctic, bands=['y_wind', 'x_wind'],
            netcdf_dim={'time':
                np.datetime64(datetime.datetime(2017,2,28,15,30,0))},
                mapperName='netcdf_cf')
        self.assertIsInstance(n, Nansat)
        self.assertEqual(2, len(n.bands()))
        self.assertTrue(n['x_wind_10m'].any())
        self.assertTrue(n['y_wind_10m'].any())

    def test_open_ecmwf_y_wind_and_x_wind_at_given_time(self):
        n = Nansat(self.test_file_ecmwf, bands=['y_wind', 'x_wind'],
                netcdf_dim={'time': '1488409200'}, mapperName='netcdf_cf')
        self.assertIsInstance(n, Nansat)
        self.assertEqual(2, len(n.bands()))
        self.assertTrue(n['x_wind_10m'].any())
        self.assertTrue(n['y_wind_10m'].any())

    def test_arome_mapper_is_used(self):
        n = Nansat(self.test_file_arome_arctic)
        self.assertEqual(n.mapper, 'arome')
        n = Nansat(self.test_file_arome_metcoop)
        self.assertEqual(n.mapper, 'arome')
        self.assertTrue(n['x_wind_10m'].any())
        self.assertTrue(n['y_wind_10m'].any())

    def test_ecmwf_mapper_is_used(self):
        n = Nansat(self.test_file_ecmwf)
        self.assertEqual(n.mapper, 'ecmwf_metno')
        self.assertTrue(n['x_wind_10m'].any())
        self.assertTrue(n['y_wind_10m'].any())

    def test_mapper_opendap_arome(self):
        n = Nansat(self.test_file_arome_opendap, mapperName='opendap_arome')
        self.assertEqual(n.mapper, 'opendap_arome')
        self.assertTrue(n['x_wind_10m'].any())
        self.assertTrue(n['y_wind_10m'].any())

    def test_reproject_arome_to_SAR(self):
        sar = Nansat(self.s1bIW)
        wind = Nansat(self.test_file_arome_metcoop, netcdf_dim={'time':
            np.datetime64(sar.time_coverage_start)},
            bands=['y_wind','x_wind'])
        self.assertTrue(wind['x_wind_10m'].any())
        self.assertTrue(wind['y_wind_10m'].any())
        wind.reproject(sar, addmask=False)
        self.assertTrue(wind['x_wind_10m'].any())
        self.assertTrue(wind['y_wind_10m'].any())

    def test_reproject_ecmwf_to_SAR(self):
        sar = Nansat(self.s1aEW)
        wind = Nansat(self.test_file_ecmwf, netcdf_dim={'time':
            np.datetime64(sar.time_coverage_start)},
            bands=['y_wind','x_wind'])
        self.assertTrue(wind['x_wind_10m'].any())
        self.assertTrue(wind['y_wind_10m'].any())
        wind.reproject(sar, addmask=False)
        self.assertTrue(wind['x_wind_10m'].any())
        self.assertTrue(wind['y_wind_10m'].any())


    def test_issue_193(self):
        fn = [
            '/vagrant/shared/test_data/cmems/GLOBAL_ANALYSIS_FORECAST_PHY_001_024-TDS-x10-X30-y55-Y73-201705181200-201705271200.nc',
            '/vagrant/shared/test_data/cmems/ARC-METNO-ARC-TOPAZ4_2_PHYS-FOR-TDS-x10-X30-y55-Y73-20170518-20170526.nc',
            '/vagrant/shared/test_data/cmems/GLOBAL_ANALYSIS_FORECAST_BIO_001_014-TDS-x-180-X179.5-y-89-Y90-20170520-20170527.nc',
        ]
        for f in fn:
            n = Nansat(f)
            self.assertTrue(n.get_metadata().has_key('time_coverage_start'))
            self.assertTrue(n.get_metadata().has_key('time_coverage_end'))
            self.assertTrue(n.get_metadata().has_key('instrument'))
            self.assertTrue(n.get_metadata().has_key('platform'))
            self.assertEqual(n.mapper, 'cmems')

if __name__ == "__main__":
    unittest.main()



# Name:         mapper_occci_online.py
# Purpose:      Nansat mapping for OC CCI data, stored online in THREDDS
# Author:       Anton Korosov
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html
import os

import json
import numpy as np
import datetime as dt

import pythesint as pti

from nansat.nsr import NSR
from nansat.mappers.opendap import Dataset, Opendap

# http://thredds.met.no/thredds/dodsC/osisaf/met.no/ice/conc/2016/04/ice_conc_sh_polstere-100_multi_201604261200.nc ice_conc
# http://thredds.met.no/thredds/dodsC/osisaf/met.no/ice/drift_lr/merged/2016/04/ice_drift_nh_polstere-625_multi-oi_201604151200-201604171200.nc dX dY
# http://thredds.met.no/thredds/dodsC/osisaf/met.no/ice/type/2016/04/ice_type_nh_polstere-100_multi_201604151200.nc ice_type
# http://thredds.met.no/thredds/dodsC/osisaf/met.no/ice/edge/2016/04/ice_edge_nh_polstere-100_multi_201604241200.nc ice_edge
# http://thredds.met.no/thredds/dodsC/osisaf_test/met.no/ice/drift_lr/merged/2013/09/ice_drift_nh_polstere-625_multi-oi_201309171200-201309191200.nc dX dY


class Mapper(Opendap):
    ''' VRT with mapping of WKV for NCEP GFS '''
    baseURLs = ['http://thredds.met.no/thredds/dodsC/cryoclim/met.no/osisaf-nh',
                'http://thredds.met.no/thredds/dodsC/osisaf_test/met.no/ice/',
                'http://thredds.met.no/thredds/dodsC/osisaf/met.no/ice/']
    timeVarName = 'time'
    xName = 'xc'
    yName = 'yc'
    t0 = dt.datetime(1978, 1, 1)
    srcDSProjection = NSR().wkt

    def __init__(self, filename, gdalDataset, gdalMetadata, date=None, ds=None, bands=None,
                 cachedir=None, **kwargs):
        ''' Create NCEP VRT
        Parameters:
            filename : URL
            date : str
                2010-05-01
            ds : netCDF.Dataset
                previously opened dataset

        '''
        self.test_mapper(filename)
        ds = Dataset(filename)
        proj4str = '%s +units=%s' % (ds.variables['Polar_Stereographic_Grid'].proj4_string,
                                     ds.variables['xc'].units)
        self.srcDSProjection = NSR(proj4str).wkt
        if filename[-3:] == '.nc':
            date = self.t0 + dt.timedelta(seconds=ds.variables['time'][0])
            date = date.strftime('%Y-%m-%d')

        self.create_vrt(filename, gdalDataset, gdalMetadata, date, ds, bands, cachedir)

        # add instrument and platform
        mm = pti.get_gcmd_instrument('Passive Remote Sensing')
        ee = pti.get_gcmd_platform('Earth Observation Satellites')
        self.dataset.SetMetadataItem('instrument', json.dumps(mm))
        self.dataset.SetMetadataItem('platform', json.dumps(ee))

    def convert_dstime_datetimes(self, dsTime):
        ''' Convert time variable to np.datetime64 '''

        dsDatetimes = np.array([np.datetime64(self.t0 + dt.timedelta(seconds=day))
                                for day in dsTime]).astype('M8[s]')
        return dsDatetimes



# Name:    node.py
# Purpose: Rapidly assemble XML using minimal coding.
# Authors:      Bruce Eckel, (c)2006 MindView Inc. www.MindView.net
# Contributors: Asuka Yamakawa, Anton Korosov, Knut-Frode Dagestad,
#               Morten W. Hansen, Alexander Myasoyedov,
#               Dmitry Petrenko, Evgeny Morozov
# Created:      29.06.2011
# Copyright:    (c) NERSC 2011 - 2013
# Licence:
# This file is part of NANSAT.
# NANSAT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
# http://www.gnu.org/licenses/gpl-3.0.html
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
from __future__ import unicode_literals

import os
import re
import xml.dom.minidom as xdm


class Node(object):
    """
    Rapidly assemble XML using minimal coding.

    By Bruce Eckel, (c)2006 MindView Inc. www.MindView.net
    Permission is granted to use or modify without payment as
    long as this copyright notice is retained.

    Everything is a Node, and each Node can either have a value
    or subnodes. Subnodes can be appended to Nodes using '+=',
    and a group of Nodes can be strung together using '+'.

    Create a node containing a value by saying
    Node('tag', 'value')
    You can also give attributes to the node in the constructor:
    Node('tag', 'value', attr1 = 'attr1', attr2 = 'attr2')
    or without a value:
    Node('tag', attr1 = 'attr1', attr2 = 'attr2')

    To produce xml from a finished Node n, say n.xml() (for
    nicely formatted output) or n.rawxml().

    You can read and modify the attributes of an xml Node using
    getAttribute(), setAttribute(), or delAttribute().

    You can find the value of the first subnode with tag == 'tag'
    by saying n['tag']. If there are multiple instances of n['tag'],
    this will only find the first one, so you should use node() or
    nodeList() to narrow your search down to a Node that only has
    one instance of n['tag'] first.

    You can replace the value of the first subnode with tag == 'tag'
    by saying n['tag'] = newValue. The same issues exist as noted
    in the above paragraph.

    You can find the first node with tag == 'tag' by saying
    node('tag'). If there are multiple nodes with the same tag
    at the same level, use nodeList('tag').

    The Node class is also designed to create a kind of 'domain
    specific language' by subclassing Node to create Node types
    specific to your problem domain.

    This implementation uses xml.dom.minidom which is available
    in the standard Python 2.4 library. However, it can be
    retargeted to use other XML libraries without much effort.

    """

    def __init__(self, tag, value=None, **attributes):
        """Everything is a Node. The XML is maintained as (very efficient)
        Python objects until an XML representation is needed.

        """
        self.tag = tag.strip()
        self.attributes = attributes
        self.children = []
        self.value = value
        if self.value:
            self.value = self.value.strip()

    def getAttribute(self, name):
        """ Read XML attribute of this node. """
        return self.attributes[name]

    def setAttribute(self, name, item):
        """ Modify XML attribute of this node. """
        self.attributes[name] = item

    def delAttribute(self, name):
        """ Remove XML attribute with this name. """
        del self.attributes[name]

    def replaceAttribute(self, name, value):
        """ replace XML arrtibute of this node. """
        del self.attributes[name]
        self.attributes[name] = value

    def node(self, tag, elemNum=0):
        """ Recursively find the first subnode with this tag.

        elemNum : int
            if there are several same tag, specify which element to take.

        """
        if self.tag == tag:
            return self
        ielm = 0
        for child in self.children:
            result = child.node(tag)
            if result and ielm == elemNum:
                return result
            elif result:
                ielm += 1
        return False

    def replaceNode(self, tag, elemNum=0, newNode=None):
        ''' Find the first subnode with this tag and replace with given node.

        tag : str
            node tag
        elemNum : int
            number of subnode among other subnodes with similar tag

        '''
        status = False
        elemi = 0
        for i in range(len(self.children)):
            if str(self.children[i].tag) == tag:
                if elemi == elemNum:
                    self.children[i] = newNode
                    status = True
                    break
                else:
                    elemi += 1
        return status

    def delNode(self, tag, options=None):
        '''
        Recursively find nodes containing subnodes with this tag and remove
        subnodes

        options : dictionary
            if there are several tags, specify a node by their attributes.

        '''
        # indeces of children to be removed
        ideleted = []

        for i, child in enumerate(self.children):
            if str(child.tag) != str(tag):
                # if child has another tag : delete children
                child.delNode(tag, options)
            elif options is None:
                # if child has this tag and no options : mark for deletion
                ideleted.append(i)
            else:
                # if child has this tag
                # and options match attributes : mark for deletion
                for key in options.keys():
                    if str(child.attributes.get(key, '')) == str(options[key]):
                        ideleted.append(i)

        # delete marked children
        for i in sorted(ideleted, reverse=True):
            self.children.pop(i)

    def find_dom_child(self, dom, tagName, n=0):
        '''Recoursively find child of the dom'''
        children = dom.childNodes
        theChild = None

        chn = 0
        for child in children:
            print(child, child.nodeType, chn)
            if child.nodeType == 1:
                print(child.tagName)
                if str(child.tagName) == tagName:
                    print(child.tagName, tagName, 'OK')
                    if chn == n:
                        theChild = child
                    chn += 1

            if theChild is not None:
                break

            if child.hasChildNodes():
                print('has childs')
                theChild = self.find_dom_child(child, tagName, n)

        return theChild

    def nodeList(self, tag):
        '''
        Produce a list of subnodes with the same tag.
        Note:
        It only makes sense to do this for the immediate
        children of a node. If you went another level down,
        the results would be ambiguous, so the user must
        choose the node to iterate over.

        '''
        return [n for n in self.children if n.tag == tag]

    def tagList(self):
        ''' Produce a list of all tags of the immediate children '''
        taglist = []
        for iTag in range(len(self.children)):
            taglist.append(str(self.children[iTag].tag))
        return taglist

    def replaceTag(self, oldTag, newTag):
        ''' Replace tag name '''
        tags = self.tagList()
        for itag, itagName in enumerate(tags):
            if itagName == oldTag:
                self.node(self.tag).children[itag].tag = newTag

    def getAttributeList(self):
        ''' get attributes and valuse from the node and return their lists '''
        nameList = []
        valList = []
        for key, val in self.attributes.items():
            nameList.append(key)
            valList.append(val)
        return nameList, valList

    def insert(self, contents):
        ''' return Node of the node with inserted <contents>'''
        dom2 = xdm.parseString(contents)
        dom1 = xdm.parseString(self.dom().toxml())
        dom1.childNodes[0].appendChild(dom1.importNode(dom2.childNodes[0],
                                                       True))
        return Node.create(dom1)

    def __getitem__(self, tag):
        '''
        Produce the value of a single subnode using operator[].
        Recursively find the first subnode with this tag.
        If you want duplicate subnodes with this tag, use
        nodeList().

        '''
        subnode = self.node(tag)
        if not subnode:
            raise KeyError
        return subnode.value

    def __setitem__(self, tag, newValue):
        '''
        Replace the value of the first subnode containing 'tag'
        with a new value, using operator[].

        '''
        assert isinstance(newValue, str), ('Value %s must be a string'
                                           % str(newValue))
        subnode = self.node(tag)
        if not subnode:
            raise KeyError
        subnode.value = newValue

    def __iadd__(self, other):
        ''' Add child nodes using operator += '''
        assert isinstance(other, Node), 'Tried to += ' + str(other)
        self.children.append(other)
        return self

    def __add__(self, other):
        ''' Allow operator + to combine children '''
        return self.__iadd__(other)

    def __str__(self):
        ''' Display this object (for debugging) '''
        result = self.tag + '\n'
        for k, v in self.attributes.items():
            result += '    attribute: %s = %s\n' % (k, v)
        if self.value:
            result += '    value: [%s]' % self.value
        return result

    # The following are the only methods that rely on the underlying
    # Implementation, and thus the only methods that need to change
    # in order to retarget to a different underlying implementation.

    # A static dom implementation object, used to create elements:
    doc = xdm.getDOMImplementation().createDocument(None, None, None)

    def dom(self):
        '''
        Lazily create a minidom from the information stored
        in this Node object.

        '''
        element = Node.doc.createElement(self.tag)
        for key, val in self.attributes.items():
            element.setAttribute(key, val)
        if self.value:
            assert not self.children, ('cannot have value and children: %s'
                                       % str(self))
            element.appendChild(Node.doc.createTextNode(self.value))
        else:
            for child in self.children:
                element.appendChild(child.dom())  # Generate children as well
        return element

    def xml(self, separator='  '):
        return self.dom().toprettyxml(separator)

    def rawxml(self):
        return str(self.dom().toxml())

    @staticmethod
    def create(dom):
        """
        Create a Node representation, given either
        a string representation of an XML doc, or a dom.

        """
        if isinstance(dom, str):
            if os.path.exists(dom):
                # parse input file
                dom = xdm.parse(dom)
            else:
                # Strip all extraneous whitespace so that
                # text input is handled consistently:
                dom = re.sub('\s+', ' ', dom)
                dom = dom.replace('> ', '>')
                dom = dom.replace(' <', '<')
                return Node.create(xdm.parseString(str(dom)))

        # To pass test for python3, decoding of bytes object is requested
        if dom.nodeType == dom.DOCUMENT_NODE:
            return Node.create(dom.childNodes[0])
        if dom.nodeName == '#text':
            return
        node = Node(dom.nodeName)
        if dom.attributes:
            for name, val in dom.attributes.items():
                node.setAttribute(name, val)
        for n in dom.childNodes:
            if n.nodeType == n.TEXT_NODE and n.wholeText.strip():
                node.value = n.wholeText
            else:
                subnode = Node.create(n)
                if subnode:
                    node += subnode
        return node



# Name:        mapper_obpg_l2
# Purpose:     Mapping for L2 data from the OBPG web-site
# Authors:      Anton Korosov
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html
from datetime import datetime, timedelta
from math import ceil
from dateutil.parser import parse

import json
import pythesint as pti

from nansat.tools import gdal, ogr
from nansat.vrt import VRT
from nansat.nsr import NSR
from nansat.mappers.obpg import OBPGL2BaseClass

from nansat.exceptions import WrongMapperError

class Mapper(OBPGL2BaseClass):
    ''' Mapper for SeaWIFS/MODIS/MERIS/VIIRS L2 data from OBPG

    TODO:
    * Test on SeaWIFS
    * Test on MODIS Terra
    '''

    def __init__(self, filename, gdalDataset, gdalMetadata,
                 GCP_COUNT=10, **kwargs):
        ''' Create VRT
        Parameters
        ----------
        GCP_COUNT : int
            number of GCPs along each dimention
        '''

        # should raise error in case of not obpg_l2 file
        try:
            title = gdalMetadata["Title"]
        except:
            raise WrongMapperError

        if title not in self.titles:
            raise WrongMapperError

        # get subdataset and parse to VRT.__init__()
        # for retrieving geo-metadata
        # but NOT from longitude or latitude because it can be smaller!
        subDatasets = gdalDataset.GetSubDatasets()
        for subDataset in subDatasets:
            if ('longitude' not in subDataset[1] and
                    'latitude' not in subDataset[1]):
                gdalSubDataset = gdal.Open(subDataset[0])
                break

        if title is 'GOCI Level-2 Data':
            # set GOCI projection parameters
            rasterXSize = 5567
            rasterYSize = 5685
            proj4 = '+proj=ortho +lat_0=36 +lon_0=130 units=m  +ellps=WGS84 +datum=WGS84 +no_defs'
            srs = osr.SpatialReference()
            srs.ImportFromProj4(proj4)
            projection = srs.ExportToWkt()
            geoTransform = (-1391500.0, 500.0, 0.0, 1349500.0, 0.0, -500.0)

            # create empty VRT dataset with georeference only
            self._init_from_dataset_params(rasterXSize, rasterYSize, geoTransform, projection)
        else:
            # create empty VRT dataset with geolocation only
            self._init_from_gdal_dataset(gdalSubDataset)

        # parts of dictionary for all Reflectances
        #dictRrs = {'wkv': 'surface_ratio_of_upwelling_radiance_emerging_from_sea_water_to_downwelling_radiative_flux_in_air', 'wavelength': '412'} }
        # dictionary for all possible bands
        allBandsDict = {'Rrs': {'src': {},
                                'dst': {'wkv': 'surface_ratio_of_upwelling_radiance_emerging_from_sea_water_to_downwelling_radiative_flux_in_air'}},
                        'Kd': {'src': {},
                               'dst': {'wkv': 'volume_attenuation_coefficient_of_downwelling_radiative_flux_in_sea_water'}},
                        'chlor_a': {'src': {},
                                    'dst': {'wkv': 'mass_concentration_of_chlorophyll_a_in_sea_water',
                                            'case': 'I'}},
                        'cdom_index': {'src': {},
                                       'dst': {'wkv': 'volume_absorption_coefficient_of_radiative_flux_in_sea_water_due_to_dissolved_organic_matter',
                                               'case': 'II'}},
                        'sst': {'src': {},
                                'dst': {'wkv': 'sea_surface_temperature'}},
                        'sst4': {'src': {},
                                 'dst': {'wkv': 'sea_surface_temperature'}},
                        'l2_flags': {'src': {'SourceType': 'SimpleSource',
                                             'DataType': 4},
                                     'dst': {'wkv': 'quality_flags',
                                             'dataType': 4}},
                        'qual_sst': {'src': {'SourceType': 'SimpleSource',
                                             'DataType': 4},
                                     'dst': {'wkv': 'quality_flags',
                                             'name': 'qual_sst',
                                             'dataType': 4}},
                        'qual_sst4': {'src': {'SourceType': 'SimpleSource',
                                              'DataType': 4},
                                      'dst': {'wkv': 'quality_flags',
                                              'name': 'qual_sst',
                                              'dataType': 4}},
                        'latitude': {'src': {},
                                     'dst': {'wkv': 'latitude'}},
                        'longitude': {'src': {},
                                      'dst': {'wkv': 'longitude'}},
                        'par': {'src': {},
                                      'dst': {'wkv': 'downwelling_photosynthetic_photon_radiance_in_sea_water'}},
                        'ipar': {'src': {},
                                      'dst': {'wkv': 'instantaneous_downwelling_photosynthetic_photon_radiance_in_sea_water'}},
                        }

        # loop through available bands and generate metaDict (non fixed)
        metaDict = []
        bandNo = 0
        for subDataset in subDatasets:
            # get sub dataset name
            subDatasetName = subDataset[1].split(' ')[1]
            self.logger.debug('Subdataset: %s' % subDataset[1])
            self.logger.debug('Subdataset name: "%s"' % subDatasetName)
            # get wavelength if applicable, get dataset name without wavelength
            try:
                wavelength = int(subDatasetName.split('_')[-1])
            except:
                wavelength = None
                subBandName = subDatasetName
            else:
                subBandName = subDatasetName.split('_')[0]

            self.logger.debug('subBandName, wavelength: %s %s' %
                              (subBandName, str(wavelength)))

            if subBandName in allBandsDict:
                # get name, slope, intercept
                self.logger.debug('name: %s' % subBandName)
                tmpSubDataset = gdal.Open(subDataset[0])
                tmpSubMetadata = tmpSubDataset.GetMetadata()
                slope = tmpSubMetadata.get('slope', '1')
                intercept = tmpSubMetadata.get('intercept', '0')
                self.logger.debug('slope, intercept: %s %s ' %
                                  (slope, intercept))
                # create meta entry
                metaEntry = {'src': {'SourceFilename': subDataset[0],
                                     'sourceBand':  1,
                                     'ScaleRatio': slope,
                                     'ScaleOffset': intercept},
                             'dst': {}
                             }
                # add more to src
                for srcKey in allBandsDict[subBandName]['src']:
                    metaEntry['src'][srcKey] = (allBandsDict[subBandName]
                                                ['src'][srcKey])
                # add dst from allBandsDict
                for dstKey in allBandsDict[subBandName]['dst']:
                    metaEntry['dst'][dstKey] = (allBandsDict[subBandName]
                                                ['dst'][dstKey])

                # add wavelength, band name to dst
                if wavelength is not None:
                    metaEntry['dst']['suffix'] = str(wavelength)
                    metaEntry['dst']['wavelength'] = str(wavelength)

                # append band metadata to metaDict
                self.logger.debug('metaEntry: %d => %s' % (bandNo,
                                                           str(metaEntry)))
                metaDict.append(metaEntry)
                bandNo += 1

                if subBandName == 'Rrs':
                    rrsSubDataset = subDataset[0]
                    metaEntryRrsw = {
                        'src': [{
                            'SourceFilename': subDataset[0],
                            'SourceBand':  1,
                            'ScaleRatio': slope,
                            'ScaleOffset': intercept,
                            'DataType': 6}],
                        'dst': {
                            'wkv': 'surface_ratio_of_upwelling_radiance_emerging_from_sea_water_to_downwelling_radiative_flux_in_water',
                            'suffix': str(wavelength),
                            'wavelength': str(wavelength),
                            'PixelFunctionType': 'NormReflectanceToRemSensReflectance',
                        }}

                    # append band metadata to metaDict
                    self.logger.debug('metaEntry: %d => %s' %
                                      (bandNo, str(metaEntryRrsw)))
                    metaDict.append(metaEntryRrsw)
                    bandNo += 1

        # add bands with metadata and corresponding values to the empty VRT
        self.create_bands(metaDict)

        # set TIME
        startYear = int(gdalMetadata['Start Year'])
        startDay = int(gdalMetadata['Start Day'])
        startMillisec = int(gdalMetadata['Start Millisec'])
        startDate = datetime(startYear, 1, 1) + timedelta(startDay-1, 0, 0,
                                                          startMillisec)
        # skip adding georeference for GOCI
        if title is 'GOCI Level-2 Data':
            return

        self._remove_geotransform()

        # add geolocation
        geoMeta = self.geolocation.data
        if len(geoMeta) > 0:
            self.dataset.SetMetadata(geoMeta, 'GEOLOCATION')

        # add GCPs
        geolocationMetadata = gdalSubDataset.GetMetadata('GEOLOCATION')
        xDatasetSource = geolocationMetadata['X_DATASET']
        xDataset = gdal.Open(xDatasetSource)

        yDatasetSource = geolocationMetadata['Y_DATASET']
        yDataset = gdal.Open(yDatasetSource)

        longitude = xDataset.ReadAsArray()
        latitude = yDataset.ReadAsArray()

        # estimate pixel/line step of the geolocation arrays
        pixelStep = int(ceil(float(gdalSubDataset.RasterXSize) /
                             float(xDataset.RasterXSize)))
        lineStep = int(ceil(float(gdalSubDataset.RasterYSize) /
                            float(xDataset.RasterYSize)))
        self.logger.debug('pixel/lineStep %f %f' % (pixelStep, lineStep))

        # ==== ADD GCPs and Pojection ====

        # estimate step of GCPs
        step0 = max(1, int(float(latitude.shape[0]) / GCP_COUNT))
        step1 = max(1, int(float(latitude.shape[1]) / GCP_COUNT))
        if str(title) == 'VIIRSN Level-2 Data':
            step0 = 64
        self.logger.debug('gcpCount: >%s<, %d %d %f %d %d',
                          title,
                          latitude.shape[0], latitude.shape[1],
                          GCP_COUNT, step0, step1)

        # generate list of GCPs
        dx = .5
        dy = .5
        gcps = []
        k = 0
        center_lon = 0
        center_lat = 0
        for i0 in range(0, latitude.shape[0], step0):
            for i1 in range(0, latitude.shape[1], step1):
                # create GCP with X,Y,pixel,line from lat/lon matrices
                lon = float(longitude[i0, i1])
                lat = float(latitude[i0, i1])
                if (lon >= -180 and lon <= 180 and lat >= -90 and lat <= 90):
                    gcp = gdal.GCP(lon, lat, 0, i1 * pixelStep + dx,
                                   i0 * lineStep + dy)
                    self.logger.debug('%d %d %d %f %f',
                                      k, gcp.GCPPixel, gcp.GCPLine,
                                      gcp.GCPX, gcp.GCPY)
                    gcps.append(gcp)
                    center_lon += gcp.GCPX
                    center_lat += gcp.GCPY
                    k += 1


        # append GCPs and lat/lon projection to the vsiDataset
        self.dataset.SetGCPs(gcps, NSR().wkt)
        self._remove_geolocation()

        # reproject GCPs
        center_lon /= k
        center_lat /= k
        srs = '+proj=stere +datum=WGS84 +ellps=WGS84 +lon_0=%f +lat_0=%f +no_defs' % (center_lon, center_lat)
        self.reproject_GCPs(srs)

        # use TPS for reprojection
        self.tps = True

        # add NansenCloud metadata
        self.dataset.SetMetadataItem('time_coverage_start',
                                     (parse(
                                      gdalMetadata['time_coverage_start']).
                                      isoformat()))
        self.dataset.SetMetadataItem('time_coverage_end',
                                     (parse(
                                      gdalMetadata['time_coverage_stop']).
                                      isoformat()))
        instrument = gdalMetadata['Sensor Name'][1:-1]
        platform = {'A': 'AQUA', 'T': 'TERRA'}[gdalMetadata['Sensor Name'][-1]]
        mm = pti.get_gcmd_instrument(instrument)
        ee = pti.get_gcmd_platform(platform)
        self.dataset.SetMetadataItem('instrument', json.dumps(mm))
        self.dataset.SetMetadataItem('platform', json.dumps(ee))
